# ---------------------------------------------------------------------
# Gufo Err: test TracebackMiddleware
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Third-party modules
import re

import pytest

# Gufo Labs modules
from gufo.err import Err

from .util import log_capture


def test_invalid_format() -> None:
    with pytest.raises(ValueError):
        Err().setup(format="unknownformat")


rx_long = re.compile(r"\|\n", re.MULTILINE)


def test_long_var() -> None:
    err = Err().setup(format="extend")
    longvar = {f"key{d}": "x" * 20 for d in range(10)}  # noqa
    try:
        msg = "trigger exception"
        raise RuntimeError(msg)
    except RuntimeError:
        with log_capture() as buffer:
            err.process()
            output = buffer.getvalue()
    assert bool(rx_long.search(output))


def test_repr_failed() -> None:
    class FailedRepr(object):
        def __repr__(self) -> str:
            msg = "Invalid repr"
            raise ValueError(msg)

    err = Err().setup(format="extend")
    instance = FailedRepr()  # noqa
    try:
        msg = "trigger exception"
        raise RuntimeError(msg)
    except RuntimeError:
        with log_capture() as buffer:
            err.process()
            output = buffer.getvalue()
    assert "instance = repr() failed: Invalid repr" in output


@pytest.mark.parametrize("fmt", ["terse", "extend"])
def test_stdin_module(fmt) -> None:
    def f():
        raise RuntimeError

    err = Err().setup(format=fmt)
    try:
        eval("f()")
    except Exception:
        with log_capture() as buffer:
            err.process()
            output = buffer.getvalue()
    assert "<stdin>" in output


def test_frame_bin_op() -> None:
    x = 1
    err = Err().setup()
    try:
        x + y
    except NameError:
        err.process()


def test_frame_bin_op2() -> None:
    x = 1
    y = "2"
    err = Err().setup()
    try:
        x + y
    except TypeError:
        err.process()


def test_frame_subscript() -> None:
    x = [1]
    err = Err().setup()
    try:
        x[2]
    except IndexError:
        err.process()


def test_frame_subscript2() -> None:
    x = [1]
    err = Err().setup()
    try:
        x[y]
    except NameError:
        err.process()



# ---------------------------------------------------------------------
# Gufo Err: TypeMatchFailFast
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""TypeMatchFailFast."""

# Python modules
from types import TracebackType
from typing import Dict, Optional, Type, Union

# Gufo Labs modules
from ..abc.failfast import BaseFailFast
from ..logger import logger


class TypeMatchFailFast(BaseFailFast):
    """Fail-fast on the given exception types and optional substrings.

    Args:
        exc:
            * Exception type.
            * string in form "module.name".
            * None to add matchers later.
        match: Optional substring matcher. Match all when set to None.
        msg: Optional message to show to log before failfast.
            `%s` will be expanded to the exception value.

    Examples:
        Single type:

        ``` py
        err.setup(fail_fast=[TypeMatchFailFast(RuntimeError)])
        ```

        Single type with match:

        ``` py
        err.setup(fail_fast=[TypeMatchFailFast(RuntimeError, match="stopped")])
        ```

        Single type with message:

        ``` py
        err.setup(fail_fast=[
            TypeMatchFailFast(RuntimeError, msg="Runtime Error: %s")
        ])
        ```

        Chaining:
        ``` py
        err.setup(
            fail_fast=[
                TypeMatchFailFast()
                .add_match(DivisionByZero)
                .add_match(ValueError, match="null value")
                .add_match(RuntimeError, msg="Runtime failure: %s")
            ]
        )
        ```
    """

    def __init__(
        self: "TypeMatchFailFast",
        exc: Optional[Union[str, Type[BaseException]]] = None,
        *,
        match: Optional[str] = None,
        msg: Optional[str] = None,
    ) -> None:
        super().__init__()
        # exc name -> match -> msg
        self.__map: Dict[str, Dict[Optional[str], Optional[str]]] = {}
        if exc:
            self.add_match(exc, match=match, msg=msg)

    def add_match(
        self: "TypeMatchFailFast",
        exc: Union[str, Type[BaseException]],
        *,
        match: Optional[str] = None,
        msg: Optional[str] = None,
    ) -> "TypeMatchFailFast":
        """Add new exception type for the fail-fast.

        Args:
            exc: Exception type or string in form "module.name".
            match: Optional substring matcher. Match all when set to None.
            msg: Optional message to show to log before failfast.
                `%s` will be expanded to the exception value.

        Returns:
            Self reference to allow chaining.

        Example:
            ``` py
            err.setup(
                fail_fast=[
                    TypeMatchFailFast()
                    .add_match(DivisionByZero)
                    .add_match(ValueError, match="null value")
                    .add_match(RuntimeError, msg="Runtime failure: %s")
                ]
            )
            ```
        """
        # Normalize to string
        xn = exc if isinstance(exc, str) else self.__exc_to_str(exc)
        # Fill mapping
        if xn not in self.__map:
            self.__map[xn] = {}
        self.__map[xn][match] = msg
        return self

    @staticmethod
    def __exc_to_str(t: Type[BaseException]) -> str:
        """Convert exception instance to string class name.

        Args:
            t: Exception type

        Returns:
            Class name in form `module.name`
        """
        return f"{t.__module__}.{t.__name__}"

    def must_die(
        self: "TypeMatchFailFast",
        t: Type[BaseException],
        v: BaseException,
        tb: TracebackType,
    ) -> bool:
        """Check if the proceess must die quickly.

        Check if exception class matches given substrings.
        """

        def msg(s: Optional[str]) -> None:
            if not s:
                return
            if "%s" in s:
                logger.critical(s, v)
            else:
                logger.critical(s)

        xn = self.__exc_to_str(t)
        chain = self.__map.get(xn)
        if not chain:
            return False
        # Try to match
        xv = str(v)
        for match in chain:
            if match and match in xv:
                msg(chain[match])
                return True
        # Wildcard match
        if None in chain:
            msg(chain[None])
            return True
        # Not matched
        return False



# ---------------------------------------------------------------------
# Gufo Err: __source_from tests
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import os
from importlib.abc import InspectLoader

# Third-party modules
import pytest

# Gufo Labs modules
from gufo.err.frame import (
    __get_source,
    __source_from_file,
    __source_from_loader,
)

SAMPLE_STR = 'SAMPLE_LINE = "this is the sample"'
SAMPLE_PATH = os.path.join("tests", "sample", "sample.py")


@pytest.fixture(scope="module")
def loader() -> InspectLoader:
    m = __import__("tests.sample.sample", {}, {}, "*")
    return m.__loader__


def is_valid_sample(s: str) -> bool:
    """Check string is valid module sample."""
    lines = s.splitlines()
    print(lines)
    if len(lines) < 9:
        return False
    return lines[8] == SAMPLE_STR


def test_source_from_file_miss() -> None:
    """Missed file."""
    assert __source_from_file("/tmp/nosuchfileanyway") is None  # noqa: S108


def test_source_from_hit() -> None:
    """Existing file."""
    source = __source_from_file(SAMPLE_PATH)
    assert source is not None
    assert is_valid_sample(source) is True


def test_source_from_loader_miss(loader: InspectLoader) -> None:
    """Missed module from loader."""
    assert __source_from_loader(loader, "xxx.nosuchmodule") is None


def test_source_from_broken_loader() -> None:
    class BrokenLoader(object): ...

    assert __source_from_loader(BrokenLoader(), "invalidmodule") is None


def test_source_from_loader_hit(loader: InspectLoader) -> None:
    source = __source_from_loader(loader, "tests.sample.sample")
    assert source is not None
    assert is_valid_sample(source) is True


def test_get_source_from_loader(loader: InspectLoader) -> None:
    source = __get_source(loader=loader, module_name="tests.sample.sample")
    assert source is not None
    assert is_valid_sample(source) is True


def test_get_source_from_file() -> None:
    source = __get_source(file_name=SAMPLE_PATH)
    assert source is not None
    assert is_valid_sample(source) is True


def test_get_source_miss(loader: InspectLoader) -> None:
    source = __get_source(
        loader=loader,
        module_name="xxx.nosuchmodule",
    )
    assert source is None



# ---------------------------------------------------------------------
# Gufo Err: Command-line utility
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""`err` utility."""

# Python module
import argparse
import datetime
import os
import re
import sys
import uuid
from dataclasses import dataclass
from enum import IntEnum
from operator import attrgetter
from typing import Callable, Dict, Iterable, List, Optional, Set

# Gufo Err modules
from . import __version__
from .codec import from_json
from .compressor import Compressor
from .formatter.loader import get_formatter
from .types import ErrorInfo


@dataclass
class ListItem(object):
    """Data structure for `err list`.

    Attributes:
        fingerprint: Stringified fingerprint.
        exception: Exception string.
        name: Application name.
        ts: Error timestamp.
        place: Error location.
    """

    fingerprint: str
    exception: str
    name: str
    ts: datetime.datetime
    place: str


class ExitCode(IntEnum):
    """Cli exit codes.

    Attributes:
        OK: Successful exit
        NOT_EXISTS: Error Info directory is not found
        EACCESS: Error Info directory is not readable
        CANNOT_READ: Cannot read Error Info file
        INVALID_ARGS: Invalid arguments
        SYNTAX: Invalid expression
    """

    OK = 0
    NOT_EXISTS = 1
    EACCESS = 2
    CANNOT_READ = 3
    INVALID_ARGS = 4
    SYNTAX = 5


class Cli(object):
    """`err` utility class."""

    rx_fn = re.compile(
        "^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}.json"
        "(|.gz|.bz2|.xz)$"
    )

    def handle_version(self: "Cli", _ns: argparse.Namespace) -> ExitCode:
        """Print Gufo Err version.

        Args:
            _ns: Options namespace, ignored.

        Returns:
            Exit code.
        """
        print(f"Gufo Err {__version__}")
        return ExitCode.OK

    @staticmethod
    def col(t: str, width: int) -> str:
        """Format text to column.

        Enlarge with spaces, when necessary.
        Cut if too long.

        Args:
            t: Text
            width: Column width

        Returns:
            Aligned text.

        Examples:
            ``` py
            col("abcdef", 5)
            ```
            returns
            ``` py
            "ab..."
            ```
        """
        ln = len(t)
        if ln < width:
            return t + " " * (width - ln)
        if ln == width:
            return t
        return t[: width - 3] + "..."

    @staticmethod
    def rcol(t: str, width: int) -> str:
        """Format text to column aligned to the range.

        Enlarge with spaces, when necessary.
        Cut if too long.

        Args:
            t: Text
            width: Column width

        Returns:
            Aligned text.

        Examples:
            ``` py
            col("abcdef", 5)
            ```
            returns
            ``` py
            "...ef"
            ```
        """
        ln = len(t)
        if ln < width:
            return t + " " * (width - ln)
        if ln == width:
            return t
        return "..." + t[ln - width - 3 :]

    @staticmethod
    def __check_dir(path: str) -> ExitCode:
        """Check if directory exists and accessible.

        Args:
            path: Directory path

        Returns:
            * OK - on success
            * NOT_EXISTS - if the directory is not exists
            * EACCESS - if the directory is not accessible
        """
        # Check if the directory exists
        if not os.path.exists(path):
            print(f"Error: {path} is not exists")
            return ExitCode.NOT_EXISTS
        # Check if the directory is readable
        if not os.access(path, os.R_OK):
            print(f"Error: {path} is not accessible")
            return ExitCode.EACCESS
        return ExitCode.OK

    def handle_list(self: "Cli", ns: argparse.Namespace) -> ExitCode:
        """Show the list of the registered errors.

        Args:
            ns: argsparse.Namespace with fields:

                * `prefix` - directory of errorinfo files.
                * `fingerprints` - list of fingerprint expressions.

        Returns:
            Exit code.
        """
        prefix = ns.prefix
        # Check if the directory exists
        code = self.__check_dir(prefix)
        if code != ExitCode.OK:
            return code
        # Resolve expressions
        fp_expr = ns.fingerprints if ns.fingerprints else ["*"]
        try:
            fingerprints = list(self.iter_fingerprints(fp_expr, prefix))
        except SyntaxError as e:
            print(f"ERROR: Invalid expression {e!s}")
            return ExitCode.SYNTAX
        # Build index
        index = self.get_index(prefix)
        # Read files
        r: List[ListItem] = []
        faults = 0
        default_ts = datetime.datetime.now()
        for fp in fingerprints:
            fn = index.get(fp)
            if not fn:
                print(f"ERROR: {fp} is not found")
                faults += 1
                continue
            info = self.read_info(os.path.join(prefix, fn))
            if not info:
                faults += 1
                continue
            # Get place
            top = info.get_app_top_frame()
            if top and top.source:
                place = f"{top.source.file_name}:{top.source.current_line}"
            else:
                place = "unknown"
            r.append(
                ListItem(
                    fingerprint=str(info.fingerprint),
                    exception=str(info.exception),
                    name=info.name,
                    ts=info.timestamp or default_ts,
                    place=place,
                )
            )
        # Get sort key
        # @todo: Make configurable
        sort_key = attrgetter("ts")
        # Print
        W_FINGER = 36
        W_EXCEPTION = 20
        W_SERVICE = 29
        W_TS = 30
        W_PLACE = 50
        print(
            " ".join(
                [
                    self.col("Fingreprint", W_FINGER),
                    self.col("Exception", W_EXCEPTION),
                    self.col("Service", W_SERVICE),
                    self.col("Time", W_TS),
                    self.col("Place", W_PLACE),
                ]
            )
        )
        print(
            " ".join(
                [
                    "-" * W_FINGER,
                    "-" * W_EXCEPTION,
                    "-" * W_SERVICE,
                    "-" * W_TS,
                    "-" * W_PLACE,
                ]
            )
        )
        for item in sorted(r, key=sort_key, reverse=False):
            print(
                " ".join(
                    [
                        self.col(str(item.fingerprint), W_FINGER),
                        self.col(item.exception, W_EXCEPTION),
                        self.col(item.name, W_SERVICE),
                        self.col(item.ts.isoformat(), W_TS),
                        self.rcol(item.place, W_PLACE),
                    ]
                )
            )
        return ExitCode.OK if not faults else ExitCode.CANNOT_READ

    @staticmethod
    def iter_fingerprints(items: List[str], prefix: str) -> Iterable[str]:
        """Resolve fingerprint expressions and iterate result.

        Fingerprint expressions is a list user-defined expressions
        passed via command line. Each item may be:

        * `<UUID>` - single fingerprint
        * `all` or `*` - all errors

        Args:
            items: List of expressions.
            prefix: Error info directory

        Returns:
            Yields all resolved fingerprints.
        """

        def is_uuid(fp: str) -> bool:
            """Check string is uuid."""
            try:
                uuid.UUID(fp)
                return True
            except ValueError:
                return False

        def resolve_all() -> Iterable[str]:
            """Resolve all fingerprints."""
            for n in os.listdir(prefix):
                fp = n.split(".")[0]
                if is_uuid(fp):
                    yield fp

        def iter_resolve(expr: str) -> Iterable[str]:
            """Resolve single expression.

            Args:
                expr: Expression

            Returns:
                Yield all resolved fingerprints.
            """
            if expr in ("all", "*"):
                yield from resolve_all()
            elif is_uuid(expr):
                # Single fingerprint
                yield expr
            else:
                # Invalid fingerprint
                raise SyntaxError(expr)

        seen: Set[str] = set()
        for expr in items:
            seen.update(iter_resolve(expr))
        yield from seen

    @staticmethod
    def get_index(prefix: str) -> Dict[str, str]:
        """Get fingerprint index.

        Args:
            prefix: Error Info directory prefix

        Returns:
            Dict of fingerprint -> file name
        """
        return {fn.split(".")[0]: fn for fn in os.listdir(prefix)}

    def handle_view(self: "Cli", ns: argparse.Namespace) -> ExitCode:
        """Show the details of the selected errors.

        Args:
            ns: argsparse.Namespace with fields:

                * `prefix` - directory of errorinfo files.
                * `format` - output format:

                    * `terse`
                    * `extend`

                * `fingerprints` - List of fingerprint expressions.

        Returns:
            Exit code.
        """
        # Get formatter
        try:
            formatter = get_formatter(ns.format)
        except ValueError:
            print(
                f"ERROR: Invalid format {ns.format}. "
                "Must be one of: terse, extend"
            )
            return ExitCode.INVALID_ARGS
        prefix = ns.prefix
        # Check if the directory exists
        code = self.__check_dir(prefix)
        if code != ExitCode.OK:
            return code
        # Resolve expressions
        try:
            fingerprints = list(
                self.iter_fingerprints(ns.fingerprints, prefix)
            )
        except SyntaxError as e:
            print(f"ERROR: Invalid expression {e!s}")
            return ExitCode.SYNTAX
        # List all files
        index = self.get_index(prefix)
        faults = 0
        for fp in fingerprints:
            path = index.get(fp)
            if not path:
                print(f"{fp} is not found")
                faults += 1
                continue
            info = self.read_info(os.path.join(prefix, path))
            if info is None:
                faults += 1
                continue
            # Format output through middleware
            print(formatter.format(info))
        return ExitCode.OK if not faults else ExitCode.CANNOT_READ

    def handle_clear(self: "Cli", ns: argparse.Namespace) -> ExitCode:
        """Clear selected errors.

        Args:
            ns: argsparse.Namespace with fields:

                * `prefix` - directory of errorinfo files.
                * `fingerprints` - List of fingerprint expressions.

        Returns:
            Exit code.
        """
        prefix = ns.prefix
        # Check if the directory exists
        code = self.__check_dir(prefix)
        if code != ExitCode.OK:
            return code
        # Resolve expressions
        try:
            fingerprints = list(
                self.iter_fingerprints(ns.fingerprints, prefix)
            )
        except SyntaxError as e:
            print(f"ERROR: Invalid expression {e!s}")
            return ExitCode.SYNTAX
        # List all files
        index = self.get_index(prefix)
        faults = 0
        for fp in fingerprints:
            path = index.get(fp)
            if not path:
                print(f"{fp} is not found")
                faults += 1
                continue
            full_path = os.path.join(prefix, path)
            try:
                os.unlink(full_path)
            except OSError as e:
                print(f"ERROR: Cannot remove file {full_path}: {e}")
                faults += 1
        return ExitCode.OK if not faults else ExitCode.CANNOT_READ

    @staticmethod
    def read_info(path: str) -> Optional[ErrorInfo]:
        """Read error info file.

        Args:
            path: JSON file path

        Returns:
          * [ErrorInfo][gufo.err.ErrorInfo] instance,
              if file has been read correctly.
          * `None` otherwise.
        """
        compressor = Compressor.autodetect(path)
        try:
            with open(path, "rb") as f:
                data = compressor.decode(f.read())
        except FileNotFoundError:
            print(f"ERROR: File {path} is not found")
            return None
        return from_json(data.decode())

    def get_handler(
        self: "Cli", name: str
    ) -> Callable[[argparse.Namespace], ExitCode]:
        """Get handler for command.

        Return the handler for furher command processing.

        Args:
            name: Command name

        Returns:
            Callable, accepting argparse.Namespace
            and returning ExitCode.
        """
        h: Callable[[argparse.Namespace], ExitCode] = getattr(
            self, f"handle_{name}"
        )
        return h

    def run(self: "Cli", args: List[str]) -> ExitCode:
        """Main dispatcher function.

        Args:
            args: List of command-line arguments.
        """
        parser = argparse.ArgumentParser(
            prog="err", description="Gufo Err reporting tool"
        )
        parser.add_argument(
            "-p",
            "--prefix",
            default=os.environ.get("GUFO_ERR_PREFIX"),
            help="JSON directory path",
        )
        subparsers = parser.add_subparsers(dest="cmd", required=True)
        # version
        subparsers.add_parser("version", help="Show Gufo Err version")
        # list
        list_parser = subparsers.add_parser(
            "list", help="Show the list of the registered errors"
        )
        list_parser.add_argument(
            "fingerprints",
            nargs=argparse.REMAINDER,
            help="Error Info fingerprint expressions",
        )
        # view
        view_parser = subparsers.add_parser("view", help="View error report")
        view_parser.add_argument(
            "-f",
            "--format",
            default="extend",
            help="Output format: terse, extend",
        )
        view_parser.add_argument(
            "fingerprints",
            nargs=argparse.REMAINDER,
            help="Error Info fingerprint expressions",
        )
        # clear
        clear_parser = subparsers.add_parser("clear", help="Remove error info")
        clear_parser.add_argument(
            "fingerprints",
            nargs=argparse.REMAINDER,
            help="Error Info fingerprint expressions",
        )
        ns = parser.parse_args(args)
        handler = self.get_handler(ns.cmd)
        return handler(ns)


def main() -> int:
    """Run err utility with command-line arguments."""
    return Cli().run(sys.argv[1:]).value



# ---------------------------------------------------------------------
# Gufo Err: TracebackMiddleware
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""TracebackMiddleware."""

# Gufo Labs modules
from ..abc.middleware import BaseMiddleware
from ..formatter.loader import get_formatter
from ..logger import logger
from ..types import ErrorInfo


class TracebackMiddleware(BaseMiddleware):
    """Dump traceback to the `gufo.err` logger.

    Args:
        format: dumping format, one of `terse`, `extend`.

    Raises:
        ValueError: On invalid `format`.

    Examples:
        Implicit initialization of the middleware using
        default `terse` format:

        ``` py
        from gufo.err import err

        err.setup()
        ```

        Implicit initialization of the middleware using
        explicit `terse` format:

        ``` py
        from gufo.err import err

        err.setup(format="terse")
        ```

        Implicit initialization of the middleware using
        explicit `extend` format:

        ``` py
        from gufo.err import err

        err.setup(format="extend")
        ```

        Explicit initialization of the middleware:

        ``` py
        from gufo.err import err
        from gufo.err.middleware.traceback import TracebackMiddleware

        err.setup(middleware=[TracebackMiddleware(format="extend")])
        ```
    """

    def __init__(
        self: "TracebackMiddleware",
        format: str = "terse",
        primary_char: str = "~",
        secondary_char: str = "^",
    ) -> None:
        super().__init__()
        self.formatter = get_formatter(
            format=format,
            primary_char=primary_char,
            secondary_char=secondary_char,
        )

    def process(self: "TracebackMiddleware", info: ErrorInfo) -> None:
        """Middleware entrypoint.

        Dumps stack info error log with given stack format.

        Args:
            info: ErrorInfo instance.
        """
        logger.error(self.formatter.format(info))



# ---------------------------------------------------------------------
# Gufo Err: ErrorInfoMiddleware
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""ErrorInfo middleware."""

# Python modules
import os
from typing import Optional

# Gufo Labs modules
from ..abc.middleware import BaseMiddleware
from ..codec import to_json
from ..compressor import Compressor
from ..logger import logger
from ..types import ErrorInfo


class ErrorInfoMiddleware(BaseMiddleware):
    """Dump error to JSON file.

    Use `err` tool to manipulate collected files.

    Args:
        path: Path to directory to write error info.
        compress: Compression algorithm. One of:

            * `None` - do not compress
            * `gz` - GZip
            * `bz2` - BZip2
            * `xz` - LZMA/xz

    Raises:
        ValueError: If path is not writable.


    Examples:
        ``` py
        from gufo.err import err

        err.setup(error_info_path="/var/err/", error_info_compress="gz")
        ```
    """

    def __init__(
        self: "ErrorInfoMiddleware", path: str, compress: Optional[str] = None
    ) -> None:
        super().__init__()
        self.path = path
        # Check permissions
        if not os.access(self.path, os.W_OK):
            msg = f"{path} is not writable"
            raise ValueError(msg)
        self.compressor = Compressor(format=compress)

    def process(self: "ErrorInfoMiddleware", info: ErrorInfo) -> None:
        """Middleware entrypoing.

        Args:
            info: ErrorInfo instance.
        """
        # ErrorInfo path
        fn = os.path.join(
            self.path, f"{info.fingerprint}.json{self.compressor.suffix}"
        )
        # Check if path is already exists
        if os.path.exists(fn):
            logger.error(
                "Error %s is already registered. Skipping.", info.fingerprint
            )
            return
        # Write erorr info
        logger.error("Writing error info into %s", fn)
        with open(fn, "wb") as f:
            f.write(self.compressor.encode(to_json(info).encode("utf-8")))



# ---------------------------------------------------------------------
# Gufo Err: TypesFailFast
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""TypesFailFast."""

# Python modules
from types import TracebackType
from typing import Iterable, Type

# Gufo Labs modules
from ..abc.failfast import BaseFailFast


class TypesFailFast(BaseFailFast):
    """Fail-fast on the given list of exception types.

    Args:
        types: Iterable of exception types.

    Examples:
        ``` py
        from gufo.err import err
        from gufo.err.types import TypesFailFast

        err.setup(fail_fast=[TypesFailFast([RuntimeError, ValueError])])
        ```
    """

    def __init__(
        self: "TypesFailFast", types: Iterable[Type[Exception]]
    ) -> None:
        super().__init__()
        self.types = set(types)

    def must_die(
        self: "TypesFailFast",
        t: Type[BaseException],
        v: BaseException,
        tb: TracebackType,
    ) -> bool:
        """Check if the process must die quickly.

        Returns true if the exception type is one of the given types.
        """
        return t in self.types



# ---------------------------------------------------------------------
# Gufo Err: err singleton
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""Define `Err` class and `err` singleton.

Attributes:
    err: Err singletone.
"""

# Python modules
import hashlib
import os
import sys
from types import TracebackType
from typing import Callable, Iterable, List, Optional, Type
from uuid import UUID

from .abc.failfast import BaseFailFast
from .abc.middleware import BaseMiddleware
from .frame import iter_frames
from .logger import logger

# Gufo Labs modules
from .types import ErrorInfo, FrameInfo

DEFAULT_NAME = "unknown"
DEFAULT_VERSION = "unknown"
DEFAULT_HASH = "sha1"
DEFAULT_EXIT_CODE = 1


class Err(object):
    """Error handling singleton.

    Example:
        ``` py
        from gufo.err import err

        err.setup()
        ```
    """

    def __init__(self: "Err") -> None:
        self.__name = DEFAULT_NAME
        self.__version = DEFAULT_VERSION
        self.__hash_fn = hashlib.sha1
        self.__initialized = False
        self.__failfast_chain: List[BaseFailFast] = []
        self.__middleware_chain: List[BaseMiddleware] = []
        self.__failfast_code = DEFAULT_EXIT_CODE
        self.__root_module: Optional[str] = None
        self.__prev_exc_hook: Optional[
            Callable[
                [Type[BaseException], BaseException, Optional[TracebackType]],
                None,
            ]
        ] = None

    def __del__(self) -> None:
        """Cleanup."""
        if self.__prev_exc_hook:
            # Restore previous exception hook
            sys.excepthook = self.__prev_exc_hook
            self.__prev_exc_hook = None

    def process(self: "Err") -> None:
        """Process current exception context in the fenced code block.

        Example:
            ``` py
            from gufo.err import err

            ...
            try:
                my_function()
            except Exception:
                err.process()
            ```
        """
        t, v, tb = sys.exc_info()
        if not t or not v or not tb:
            return  # Not an exception context
        self.__process(t, v, tb)

    def __process(
        self: "Err",
        t: Type[BaseException],
        v: BaseException,
        tb: Optional[TracebackType] = None,
    ) -> None:
        """Process given exception context.

        Called either from .process()
        or as sys.excepthook for unhandled exceptions.

        Args:
            t: Exception type.
            v: Exception value.
            tb: Traceback frame.

        Raises:
            RuntimeError: If setup() is not called.
        """
        if not self.__initialized:
            msg = "setup() is not called"
            raise RuntimeError(msg)
        if not tb:
            return
        if t in (SystemExit, KeyboardInterrupt):
            raise  # noqa: PLE0704 Do not mess the exit sequence
        if self.__must_die(t, v, tb):
            os._exit(self.__failfast_code)  # Fatal error, die quickly
        # Collect stack frames
        # @todo: separate handling of endless recursion
        stack = list(iter_frames(tb))
        # Calculate error fingerprint
        fp = self.__fingerprint(t, v, stack)
        # Build stack info
        err_info = ErrorInfo(
            name=self.__name,
            version=self.__version,
            fingerprint=fp,
            stack=stack,
            exception=v,
            root_module=self.__root_module,
        )
        # Process the response
        self.__run_middleware(err_info)

    def setup(
        self: "Err",
        *,
        catch_all: bool = False,
        root_module: Optional[str] = None,
        name: str = DEFAULT_NAME,
        version: str = DEFAULT_VERSION,
        hash: str = DEFAULT_HASH,
        fail_fast: Optional[Iterable[BaseFailFast]] = None,
        fail_fast_code: int = DEFAULT_EXIT_CODE,
        middleware: Optional[Iterable[BaseMiddleware]] = None,
        format: Optional[str] = "terse",
        error_info_path: Optional[str] = None,
        error_info_compress: Optional[str] = None,
    ) -> "Err":
        """Setup error handling singleton.

        Must be called only once.

        Args:
            catch_all: Install global system exception hook.
            name: Application or service name.
            version: Application or service version.
            root_module: Top-level application module/namespace for
                split stack fingerprinting. Topmost frame from the root or
                the nested modules will be considered in the error
                fingerprint.
            hash: Fingerprint hashing function name. Available functions
                are: sha256, sha3_512, blake2s, sha3_224, md5, sha384,
                sha3_256, shake_256, blake2b, sha224, shake_128, sha3_384,
                sha1, sha512. Refer to the Python's hashlib for details.
            fail_fast: Iterable of BaseFailFast instances for fail-fast
                detection.
                Process will terminate with `fail_fast_code` error code
                if any of instances in the chain will return True.
            fail_fast_code: System exit code on fail-fast termination.
            middleware: Iterable of BaseMiddleware instances
                for error processing middleware.
                Instances are evaluated in the order of appearance.
            format: If not None install TracebackMiddleware for given
                output format.
            error_info_path: If not None install ErrorInfoMiddleware.
                `error_info_path` should point to a writable directories,
                in which the error info files to be written.
            error_info_compress: Used only with `error_info_path`.
                Set error info compression method. One of:

                * `None` - do not compress
                * `gz` - GZip
                * `bz2` - BZip2
                * `xz` - LZMA/xz

        Returns:
            Err instance.

        Raises:
            RuntimeError: When called twice.
            ValueError: On configuration parameters error.
        """
        if self.__initialized:
            msg = "Already initialized"
            raise RuntimeError(msg)
        # Install system-wide exception hook
        if catch_all:
            self.__prev_exc_hook = sys.excepthook
            sys.excepthook = self.__process
        # Init parameters
        self.__name = name or DEFAULT_NAME
        self.__version = version or DEFAULT_VERSION
        self.__failfast_code = fail_fast_code
        self.__root_module = root_module
        try:
            self.__hash_fn = getattr(hashlib, hash)
        except AttributeError as e:
            msg = f"Unknown hash: {hash}"
            raise ValueError(msg) from e
        # Initialize fail fast chain
        if fail_fast:
            self.__failfast_chain = []
            for ff in fail_fast:
                self.add_fail_fast(ff)
        else:
            self.__failfast_chain = []
        # Initialize response chain
        if middleware:
            self.__middleware_chain = self.__default_middleware(
                format=format,
                error_info_path=error_info_path,
                error_info_compress=error_info_compress,
            )
            for resp in middleware:
                self.add_middleware(resp)
        else:
            self.__middleware_chain = self.__default_middleware(
                format=format,
                error_info_path=error_info_path,
                error_info_compress=error_info_compress,
            )
        # Mark as initialized
        self.__initialized = True
        return self

    def __must_die(
        self: "Err",
        t: Type[BaseException],
        v: BaseException,
        tb: TracebackType,
    ) -> bool:
        """Check if the error is fatal and the process must die.

        Process fail-fast sequence and return True if the
        process must die quickly.
        """
        if not tb:
            return False
        return any(ff.must_die(t, v, tb) for ff in self.__failfast_chain)

    def __run_middleware(self: "Err", err_info: ErrorInfo) -> None:
        """Process all the middleware.

        Args:
            err_info: Filled ErrorInfo structure
        """
        for resp in self.__middleware_chain:
            try:
                resp.process(err_info)
            except Exception as e:  # noqa: BLE001
                logger.error("%r middleware failed: %s", resp, e)

    def iter_fingerprint_parts(
        self: "Err",
        t: Type[BaseException],
        v: BaseException,
        stack: List[FrameInfo],
    ) -> Iterable[str]:
        """Iterate over the fingerprint parts.

        Iterable to yield all fingerprint parts.
        May be overriden in subclasses.

        Args:
            t: Exception type.
            v: Exception instance:
            stack: Current stack.

        Returns:
            Iterable of strings.
        """
        yield self.__name  # Service name
        yield self.__version  # Service version
        yield t.__name__  # Exception class
        # Top-level stack info
        if stack:
            top = stack[0]
            yield top.module or "unknown"  # Top module
            yield top.name  # Top callable name
            if top.source:
                yield str(top.source.current_line)  # Top execution line
        # Application stack info
        if self.__root_module:
            app_top = None
            prefix = f"{self.__root_module}."
            for frame in stack:
                if frame.module and (
                    frame.module == self.__root_module
                    or frame.module.startswith(prefix)
                ):
                    app_top = frame
                    break
            if app_top:
                yield app_top.module or "unknown"  # App module
                yield app_top.name  # App module Current callable name
                if app_top.source:
                    yield str(
                        app_top.source.current_line
                    )  # App execution line

    def __fingerprint(
        self: "Err",
        t: Type[BaseException],
        v: BaseException,
        stack: List[FrameInfo],
    ) -> UUID:
        """Calculate the error fingerprint.

        Calculate error fingerprint for given exception
        and the stack. Fingerprint is stable for repeating
        conditions.

        Args:
            t: Exception type.
            v: Exception instance:
            stack: Current stack.

        Returns:
            Error fingerprint as UUID.
        """
        fp_hash = self.__hash_fn(
            b"\x00".join(
                x.encode("utf-8")
                for x in self.iter_fingerprint_parts(t, v, stack)
            )
        ).digest()
        return UUID(bytes=fp_hash[:16], version=5)

    def add_fail_fast(self: "Err", ff: BaseFailFast) -> None:
        """Add fail-fast handler to the end of the chain.

        Args:
            ff: BaseFailFast instance.

        Raises:
            ValueError: If `ff` is not BaseFailFast instance.
        """
        if not isinstance(ff, BaseFailFast):
            msg = "add_fail_fast() argument must be BaseFailFast instance"
            raise ValueError(msg)
        self.__failfast_chain.append(ff)

    def add_middleware(self: "Err", mw: BaseMiddleware) -> None:
        """Add middleware to the end of the chain.

        Args:
            mw: BaseMiddleware instance

        Raises:
            ValueError: If `mw` is not BaseMiddleware instance.
        """
        if not isinstance(mw, BaseMiddleware):
            msg = "add_response() argument must be BaseMiddleware instance"
            raise ValueError(msg)
        self.__middleware_chain.append(mw)

    def __default_middleware(
        self: "Err",
        format: Optional[str] = None,
        error_info_path: Optional[str] = None,
        error_info_compress: Optional[str] = None,
    ) -> List[BaseMiddleware]:
        """Get default middleware chain.

        Args:
            format: traceback format. See TracebackMiddleware for details.
                Do not configure tracebacks if None.
            error_info_path: Directory path to write error info.
                See ErrorInfoMiddleware for details.
                Do not configure middleware if None.
            error_info_compress: Error info compression algorithm. Used along
                with `error_info_path`.
        """
        r: List[BaseMiddleware] = []
        if format is not None:
            from .middleware.traceback import TracebackMiddleware

            r.append(TracebackMiddleware(format=format))
        if error_info_path is not None:
            from .middleware.errorinfo import ErrorInfoMiddleware

            r.append(
                ErrorInfoMiddleware(
                    path=error_info_path, compress=error_info_compress
                )
            )
        return r


# Define the singleton
err = Err()



# ---------------------------------------------------------------------
# Gufo Err: Frame Extraction
# ---------------------------------------------------------------------
# Copyright (C) 2022-25, Gufo Labs
# ---------------------------------------------------------------------
"""FrameInfo structure."""

# Python modules
import ast
import sys
from importlib.abc import InspectLoader
from itertools import islice
from types import CodeType, TracebackType
from typing import Iterable, Optional, cast

# Gufo Labs modules
from .types import Anchor, CodePosition, FrameInfo, SourceInfo

PY_3 = 3
PY_3_11 = (3, 11)


def exc_traceback() -> TracebackType:
    """Cast type to `sys.exc_info()`.

    Extract and return top-level excecution frame
    from current exception context.

    Returns:
        Top-level exception frame.
    """
    return cast(TracebackType, sys.exc_info()[2])


def iter_frames(
    tb: TracebackType, context_lines: int = 7
) -> Iterable[FrameInfo]:
    """Iterate over traceback frames.

    Args:
        tb: current execution frame.
        context_lines: Source code context to extract.
            Current line, up to `context_lines` below
            the current line, and up to `context_lines`
            above the current line will be extracted.

    Returns:
        Iterable of FrameInfo, starting from top of the
        stack (current code position).
    """
    current: Optional[TracebackType] = tb
    while current is not None:
        frame = current.tb_frame
        src = __get_source_info(
            file_name=frame.f_code.co_filename,
            line_no=current.tb_lineno,
            context_lines=context_lines,
            loader=frame.f_globals.get("__loader__"),
            module_name=frame.f_globals.get("__name__"),
            code=frame.f_code,
            inst_index=current.tb_lasti,
        )
        yield FrameInfo(
            name=frame.f_code.co_name,
            module=frame.f_globals.get("__name__"),
            source=src,
            locals=frame.f_locals,
        )
        current = current.tb_next


def __source_from_loader(
    loader: InspectLoader, module_name: str
) -> Optional[str]:
    try:
        return loader.get_source(module_name)
    except AttributeError:
        return None  # .get_source() does not supported
    except ImportError:
        return None


def __source_from_file(file_name: str) -> Optional[str]:
    try:
        with open(file_name) as f:
            return f.read()
    except OSError:
        return None


def __get_source(
    file_name: Optional[str] = None,
    loader: Optional[InspectLoader] = None,
    module_name: Optional[str] = None,
) -> Optional[str]:
    src: Optional[str] = None
    if loader and module_name:
        src = __source_from_loader(loader, module_name)
    if not src and file_name:
        src = __source_from_file(file_name)
    return src


def __get_source_info(
    line_no: int,
    context_lines: int,
    code: CodeType,
    inst_index: int,
    file_name: Optional[str] = None,
    loader: Optional[InspectLoader] = None,
    module_name: Optional[str] = None,
) -> Optional[SourceInfo]:
    src = __get_source(
        file_name=file_name, loader=loader, module_name=module_name
    )
    if not src:
        return None  # Unable to get the source
    # @todo: maybe use linecache
    lines = src.splitlines()  # @todo: Implement sliding line iterator
    # Extract current code position
    code_position = __get_code_position(code, inst_index, lines[line_no - 1])
    if code_position:
        # Exact locations
        first_line = max(1, code_position.start_line - context_lines)
        last_line = code_position.end_line + context_lines
    else:
        first_line = max(1, line_no - context_lines)
        last_line = line_no + context_lines
    return SourceInfo(
        file_name=file_name or module_name or "",
        first_line=first_line,
        current_line=line_no,
        lines=lines[first_line - 1 : last_line],
        pos=code_position,
    )


def __has_code_position() -> bool:
    """Check if python supports exact code positions.

    Returns:
        * True - if python 3.11+ and PYTHONNODEBUGRANGES is not set.
        * False - otherwise
    """
    import os

    if sys.version_info.major < PY_3:
        return False  # No support for Python 2
    if sys.version_info.major > PY_3:
        return True  # Python 4? :)
    if sys.version_info >= PY_3_11:
        return os.environ.get("PYTHONNODEBUGRANGES") is None
    return False


HAS_CODE_POSITION = __has_code_position()


def __get_code_position(
    code: CodeType, inst_index: int, line: str
) -> Optional[CodePosition]:
    """Extract code range for current instruction.

    Args:
        code: Code object
        inst_index: Current instruction index, usually from `tb_lasti`
        line: Current code line

    Returns:
        Optional CodePosition instance
    """
    if not HAS_CODE_POSITION or inst_index < 0:
        return None
    # Warning! co_positions is not defineed prior the Python 3.11
    # so mypy will raise an error.
    positions_gen = code.co_positions()
    start_line, end_line, start_col, end_col = next(
        islice(positions_gen, inst_index // 2, None)
    )
    if (
        start_line is None
        or end_line is None
        or start_col is None
        or end_col is None
    ):
        return None
    if start_line == end_line:
        anchor = __get_anchor(line[start_col:end_col], start_col)
    else:
        anchor = None
    return CodePosition(
        start_line=start_line,
        end_line=end_line,
        start_col=start_col,
        end_col=end_col,
        anchor=anchor,
    )


def __get_anchor(segment: str, indent: int = 0) -> Optional[Anchor]:
    """Split code segment and try to get error anchors.

    Backport from Python 3.11
    `_extract_caret_anchors_from_line_segment`.

    Args:
        segment: Code segment with current op.
        indent: Position offset.

    Returns:
        * Anchor instance if code can be refined.
        * None otherwise
    """
    try:
        tree = ast.parse(segment)
    except SyntaxError:
        return None
    # Should be only current statement
    if len(tree.body) != 1:
        return None
    statement = tree.body[0]
    # Cannot use match as it raises SyntaxError
    # prior Python 3.10
    if not isinstance(statement, ast.Expr):
        return None
    expr = statement.value
    if isinstance(expr, ast.BinOp):
        # Binary operation, problem with operator
        operator_str = segment[
            expr.left.end_col_offset : expr.right.col_offset
        ]
        operator_offset = len(operator_str) - len(operator_str.lstrip())
        if expr.left.end_col_offset is None:
            return None
        left_anchor = expr.left.end_col_offset + operator_offset
        right_anchor = left_anchor + 1
        if (
            operator_offset + 1 < len(operator_str)
            and not operator_str[operator_offset + 1].isspace()
        ):
            right_anchor += 1
        return Anchor(left=left_anchor + indent, right=right_anchor + indent)
    if (
        isinstance(expr, ast.Subscript)
        and expr.value.end_col_offset is not None
        and expr.slice.end_col_offset is not None
    ):
        # Subscript operation, problem with value
        return Anchor(
            left=expr.value.end_col_offset + indent,
            right=expr.slice.end_col_offset + indent + 1,
        )
    return None



# ---------------------------------------------------------------------
# Gufo Err: serde tests
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import datetime
import os
import uuid
from typing import Any, Dict

# Third-party modules
import pytest
from gufo.err.codec import (
    ExceptionStub,
    from_dict,
    from_json,
    to_dict,
    to_json,
)

# Gufo Labs modules
from gufo.err.types import ErrorInfo, FrameInfo, SourceInfo

TZ = datetime.timezone(datetime.timedelta(hours=1), "CEST")

SAMPLE = ErrorInfo(
    name="oops",
    version="1.0",
    fingerprint=uuid.UUID("be8ccd86-3661-434c-8569-40dd65d9860a"),
    exception=RuntimeError("oops"),
    timestamp=datetime.datetime(2022, 3, 22, 7, 21, 29, 215827, tzinfo=TZ),
    root_module="tests",
    stack=[
        FrameInfo(
            name="test_iter_frames",
            module="tests.test_frames",
            source=SourceInfo(
                file_name=os.path.join("tests", "test_frames.py"),
                current_line=125,
                first_line=118,
                lines=[
                    "",
                    "",
                    "def test_iter_frames():",
                    '    """',
                    "    Call the function which raises an exception",
                    '    """',
                    "    try:",
                    "        entry()",
                    '        assert False, "No trace"',
                    "    except RuntimeError:",
                    "        frames = list(iter_frames(exc_traceback()))",
                    "        assert frames == SAMPLE_FRAMES",
                ],
            ),
            locals={},
        ),
        FrameInfo(
            name="entry",
            module="tests.sample.trace",
            source=SourceInfo(
                file_name=os.path.join("tests", "sample", "trace.py"),
                current_line=14,
                first_line=7,
                lines=[
                    "    x += 1",
                    "    oops()",
                    "",
                    "",
                    "def entry():",
                    "    s = 2",
                    "    s += 1",
                    "    to_oops()",
                ],
            ),
            locals={"s": 3},
        ),
        FrameInfo(
            name="to_oops",
            module="tests.sample.trace",
            source=SourceInfo(
                file_name=os.path.join("tests", "sample", "trace.py"),
                current_line=8,
                first_line=1,
                lines=[
                    "def oops():",
                    '    raise RuntimeError("oops")',
                    "",
                    "",
                    "def to_oops():",
                    "    x = 1",
                    "    x += 1",
                    "    oops()",
                    "",
                    "",
                    "def entry():",
                    "    s = 2",
                    "    s += 1",
                    "    to_oops()",
                ],
            ),
            locals={"x": 2},
        ),
        FrameInfo(
            name="oops",
            module="tests.sample.trace",
            source=SourceInfo(
                file_name=os.path.join("tests", "sample", "trace.py"),
                current_line=2,
                first_line=1,
                lines=[
                    "def oops():",
                    '    raise RuntimeError("oops")',
                    "",
                    "",
                    "def to_oops():",
                    "    x = 1",
                    "    x += 1",
                    "    oops()",
                    "",
                ],
            ),
            locals={},
        ),
    ],
)

SAMPLE_DICT: Dict[str, Any] = {
    "$type": "errorinfo",
    "$version": "1.0",
    "fingerprint": "be8ccd86-3661-434c-8569-40dd65d9860a",
    "name": "oops",
    "timestamp": "2022-03-22T07:21:29.215827+01:00",
    "version": "1.0",
    "root_module": "tests",
    "exception": {"class": "RuntimeError", "args": ["oops"]},
    "stack": [
        {
            "name": "test_iter_frames",
            "module": "tests.test_frames",
            "locals": {},
            "source": {
                "file_name": "tests/test_frames.py",
                "current_line": 125,
                "first_line": 118,
                "lines": [
                    "",
                    "",
                    "def test_iter_frames():",
                    '    """',
                    "    Call the function which raises an exception",
                    '    """',
                    "    try:",
                    "        entry()",
                    '        assert False, "No trace"',
                    "    except RuntimeError:",
                    "        frames = list(iter_frames(exc_traceback()))",
                    "        assert frames == SAMPLE_FRAMES",
                ],
            },
        },
        {
            "name": "entry",
            "module": "tests.sample.trace",
            "locals": {"s": 3},
            "source": {
                "file_name": "tests/sample/trace.py",
                "current_line": 14,
                "first_line": 7,
                "lines": [
                    "    x += 1",
                    "    oops()",
                    "",
                    "",
                    "def entry():",
                    "    s = 2",
                    "    s += 1",
                    "    to_oops()",
                ],
            },
        },
        {
            "name": "to_oops",
            "module": "tests.sample.trace",
            "locals": {"x": 2},
            "source": {
                "file_name": "tests/sample/trace.py",
                "current_line": 8,
                "first_line": 1,
                "lines": [
                    "def oops():",
                    '    raise RuntimeError("oops")',
                    "",
                    "",
                    "def to_oops():",
                    "    x = 1",
                    "    x += 1",
                    "    oops()",
                    "",
                    "",
                    "def entry():",
                    "    s = 2",
                    "    s += 1",
                    "    to_oops()",
                ],
            },
        },
        {
            "name": "oops",
            "module": "tests.sample.trace",
            "locals": {},
            "source": {
                "file_name": "tests/sample/trace.py",
                "current_line": 2,
                "first_line": 1,
                "lines": [
                    "def oops():",
                    '    raise RuntimeError("oops")',
                    "",
                    "",
                    "def to_oops():",
                    "    x = 1",
                    "    x += 1",
                    "    oops()",
                    "",
                ],
            },
        },
    ],
}


def test_to_dict():
    out = to_dict(SAMPLE)
    assert out == SAMPLE_DICT


def test_to_json():
    out = to_json(SAMPLE)
    isinstance(out, str)


def test_from_dict():
    out = from_dict(SAMPLE_DICT)
    assert isinstance(out.exception, ExceptionStub)
    assert out.exception.kls == "RuntimeError"
    assert out.exception.args == ("oops",)
    out.exception = SAMPLE.exception
    assert out == SAMPLE


def test_from_json():
    src = to_json(SAMPLE)
    out = from_json(src)
    assert isinstance(out.exception, ExceptionStub)
    assert out.exception.kls == "RuntimeError"
    assert out.exception.args == ("oops",)
    out.exception = SAMPLE.exception
    assert out == SAMPLE


def test_from_dict_no_dict():
    with pytest.raises(ValueError):
        from_dict(None)


def test_from_dict_invalid_type():
    with pytest.raises(ValueError):
        from_dict({"$type": "foobar"})


def test_from_dict_invalid_version():
    with pytest.raises(ValueError):
        from_dict({"$type": "errorinfo", "$version": "-1"})


def test_from_dict_no_exc():
    with pytest.raises(ValueError):
        from_dict({"$type": "errorinfo", "$version": "1.0"})



# ---------------------------------------------------------------------
# Gufo Err: middleware module
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""Middleware handlers.

Configured middleware handlers are launched on every unhandled exception
and allows to customize an error processing. See
[BaseMiddleware][gufo.err.abc.middleware.BaseMiddleware] for details.

Available out-of-box:

* [ErrorInfoMiddleware][gufo.err.middleware.errorinfo.ErrorInfoMiddleware]:
  Dump errors to JSON files.
* [SentryMiddleware][gufo.err.middleware.sentry.SentryMiddleware]:
  Sentry integration.
* [TracebackMiddleware][gufo.err.middleware.traceback.TracebackMiddleware]:
  Dump traceback to the gufo.err logger.
"""



# ---------------------------------------------------------------------
# Gufo Err: SentryMiddleware tests
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Gufo Labs modules
from gufo.err import Err
from gufo.err.middleware.sentry import SentryMiddleware


def test_sentry_middleware():
    # Configure sentry without URL
    err = Err().setup(
        middleware=[
            SentryMiddleware("http://public@127.0.0.1:9999/1", debug=True)
        ]
    )
    # Without __before_send
    try:
        raise RuntimeError()
    except RuntimeError:
        err.process()


def test_sentry_middleware_skip_before_send():
    # Configure sentry without URL
    err = Err().setup(
        middleware=[
            SentryMiddleware(
                "http://public@127.0.0.1:9999/1",
                debug=True,
                before_send=lambda _event, _hint: None,
            )
        ]
    )
    # Without __before_send
    try:
        raise RuntimeError()
    except RuntimeError:
        err.process()



# ---------------------------------------------------------------------
# Gufo Err: SentryMiddleware
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""Sentry integration middleware."""

# Python modules
from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, Optional

# Third-party modules
import sentry_sdk

# Gufo Labs Modules
from ..abc.middleware import BaseMiddleware
from ..types import ErrorInfo

_current_err = ContextVar[Optional[ErrorInfo]]("current_err", default=None)

if TYPE_CHECKING:
    Event = sentry_sdk._types.Event
else:
    Event = Dict[str, Any]


@contextmanager
def _err_context(info: ErrorInfo) -> Generator[None, None, None]:
    """Context manager to set `__current_err` context variable.

    Set __current_err context variable with error info
    and remove it on exit.
    """
    token = _current_err.set(info)
    yield
    _current_err.reset(token)


class SentryMiddleware(BaseMiddleware):
    """[Sentry](https://sentry.io/) integration.

    `SentryMiddleware` is the wrapper around sentry_sdk
    to seamless integration into Gufo Err.

    Args:
        dsn: URL of the sentry installation. If not provided,
            use `SENTRY_DSN` envoronment variable.
            Do not send any events if value is not set.
        debug: Turns debug mode on and off.
        release: Set current version explicitly. Sentry
            will try to determine the version automatically
            if not set.
        before_send: The function accepting sentry event
            object and returning modified event object or
            nothing to stop processing.

    Examples:
        To configure the SentryMiddleware:

        ``` py
        from gufo.err import err
        from gufo.err.middleware.sentry import SentryMiddleware

        err.setup()
        err.add_middleware(
            SentryMiddleware("http://127.0.0.1:1000/", debug=True)
        )
        ```
    """

    def __init__(
        self: "SentryMiddleware",
        dsn: Optional[str] = None,
        debug: bool = False,
        release: Optional[str] = None,
        before_send: Optional[
            Callable[
                [Event, Dict[str, Any]],
                Optional[Event],
            ]
        ] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__()
        self.__user_before_send = before_send
        sentry_sdk.init(
            dsn,
            debug=debug,
            release=release,
            before_send=self.__before_send,
            **kwargs,  # type: ignore
            # shutdown_timeout=config.sentry.shutdown_timeout,
            # release=version.version,
            # max_breadcrumbs=config.sentry.max_breadcrumbs,
            # default_integrations=config.sentry.default_integrations,
            # before_send=before_send,
        )

    def __before_send(
        self: "SentryMiddleware",
        event: Event,
        hint: Dict[str, Any],
    ) -> Optional[Event]:
        """Enrich event with user-defined information and fingerprint.

        Call user-defined `before_send` and add additional
        fingerprint information.

        Args:
            event: Sentry event object.
            hint: Error hints.

        Returns:
            Event object
        """
        # Call user-defined handler
        if (
            self.__user_before_send
            and self.__user_before_send(event, hint) is None
        ):
            return None  # User handler interrupts processing

        if "exc_info" not in hint:
            return event

        exception = hint["exc_info"][1]
        info = _current_err.get()
        event["fingerprint"] = [
            "{{ type }}",
            str(exception),
            str(info.fingerprint) if info else "",
        ]
        return event

    def process(self: "SentryMiddleware", info: ErrorInfo) -> None:
        """Middleware entrypoint.

        Args:
            info: ErrorInfo instance.
        """
        with _err_context(info):
            sentry_sdk.capture_exception()



# ---------------------------------------------------------------------
# Gufo Err: Types
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""Public API data types."""

# Python modules
import datetime
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple
from uuid import UUID


@dataclass
class Anchor(object):
    """Exact problem position (Python 3.11+).

    Denotes operator of subscript which causes the problem.

    Args:
        left: Starting column.
        right: Stopping column.
    """

    left: int
    right: int


@dataclass
class CodePosition(object):
    """Exact code position for Python 3.11+.

    Args:
        start_line: First line of code
        end_line: Last line of code
        start_col: Starting column (on start_line)
        end_col: Ending column (on end_line)
        anchor: Problem anchor
    """

    start_line: int
    end_line: int
    start_col: int
    end_col: int
    anchor: Optional[Anchor]


@dataclass
class SourceInfo(object):
    """Source context for frame.

    Args:
        file_name: Normalized file name.
        first_line: first line of source context.
        current_line: current execution line.
        lines: List of lines, starting from `first_line`
        pos: Optional exact code position for Python 3.11+
    """

    file_name: str
    first_line: int
    current_line: int
    lines: List[str]
    pos: Optional[CodePosition] = None


@dataclass
class FrameInfo(object):
    """Execution frame.

    Args:
        name: Current callable name.
        source: Optional `SourceInfo` procedure. May be missed
            on loader problems.
        locals: Dicts of local variables.
        module: Python module name.
    """

    name: str
    source: Optional[SourceInfo]
    locals: Dict[str, Any]
    module: Optional[str] = None


@dataclass
class ErrorInfo(object):
    """Current execution frame information.

    Args:
        name: Application or service name, as set by
            [setup()][gufo.err.Err.setup]
        version: Application or service version, as set by
            [setup()][gufo.err.Err.setup]
        fingerprint: Error fingerprint.
        stack: List of `FrameInfo`. Current execution frame is first.
        exception: Exception instance, if caught.
        timestamp: Error timestamp.
        root_module: Optional root module, as set by
            [setup()][gufo.err.Err.setup]
    """

    name: str
    version: str
    fingerprint: UUID
    stack: List[FrameInfo]
    exception: BaseException
    timestamp: Optional[datetime.datetime] = None
    root_module: Optional[str] = None

    def get_app_top_frame(self: "ErrorInfo") -> Optional[FrameInfo]:
        """Get application's top stack frame.

        Find top stack frame belonging to the application,
        if `root_module` is set, or return stack top otherwise.

        Returns:
            * [FrameInfo][gufo.err.FrameInfo] if the stack is not empty.
            * None otherwise.
        """
        if not self.stack:
            return None
        if self.root_module:
            prefix = f"{self.root_module}."
            for frame in self.stack:
                if frame.module and (
                    frame.module == self.root_module
                    or frame.module.startswith(prefix)
                ):
                    return frame
        return self.stack[0]


class ExceptionStub(Exception):
    """Stub to deserialized exceptions.

    Args:
        kls: Exception class name
        args: Exception arguments
    """

    def __init__(
        self: "ExceptionStub", kls: str, args: Tuple[Any, ...]
    ) -> None:
        self.kls = kls
        self.args = args

    def __str__(self: "ExceptionStub") -> str:
        """Format exception to string.

        Returns:
            Formatted string.
        """
        if not self.args:
            return self.kls
        return f"{self.kls}: {self.args[0]}"



# ---------------------------------------------------------------------
# Gufo Err: test types
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import os
import uuid

# Gufo Err modules
from gufo.err import ErrorInfo, FrameInfo, SourceInfo

SAMPLE_STACK = [
    FrameInfo(
        name="test_iter_frames",
        module="tests.test_frames",
        source=SourceInfo(
            file_name=os.path.join("tests", "test_frames.py"),
            current_line=125,
            first_line=118,
            lines=[
                "",
                "",
                "def test_iter_frames():",
                '    """',
                "    Call the function which raises an exception",
                '    """',
                "    try:",
                "        entry()",
                '        assert False, "No trace"',
                "    except RuntimeError:",
                "        frames = list(iter_frames(exc_traceback()))",
                "        assert frames == SAMPLE_FRAMES",
            ],
        ),
        locals={},
    ),
    FrameInfo(
        name="entry",
        module="tests.sample.trace",
        source=SourceInfo(
            file_name=os.path.join("tests", "sample", "trace.py"),
            current_line=14,
            first_line=7,
            lines=[
                "    x += 1",
                "    oops()",
                "",
                "",
                "def entry():",
                "    s = 2",
                "    s += 1",
                "    to_oops()",
            ],
        ),
        locals={"s": 3},
    ),
    FrameInfo(
        name="to_oops",
        module="tests.sample.trace",
        source=SourceInfo(
            file_name=os.path.join("tests", "sample", "trace.py"),
            current_line=8,
            first_line=1,
            lines=[
                "def oops():",
                '    raise RuntimeError("oops")',
                "",
                "",
                "def to_oops():",
                "    x = 1",
                "    x += 1",
                "    oops()",
                "",
                "",
                "def entry():",
                "    s = 2",
                "    s += 1",
                "    to_oops()",
            ],
        ),
        locals={"x": 2},
    ),
    FrameInfo(
        name="oops",
        module="tests.sample.trace",
        source=SourceInfo(
            file_name=os.path.join("tests", "sample", "trace.py"),
            current_line=2,
            first_line=1,
            lines=[
                "def oops():",
                '    raise RuntimeError("oops")',
                "",
                "",
                "def to_oops():",
                "    x = 1",
                "    x += 1",
                "    oops()",
                "",
            ],
        ),
        locals={},
    ),
]


def test_app_top_empty_stack():
    info = ErrorInfo(
        name="unknown",
        version="unknown",
        fingerprint=uuid.uuid4(),
        stack=[],
        exception=ValueError(),
    )
    top = info.get_app_top_frame()
    assert top is None


def test_app_top_no_root_module():
    info = ErrorInfo(
        name="unknown",
        version="unknown",
        fingerprint=uuid.uuid4(),
        stack=SAMPLE_STACK,
        exception=ValueError(),
    )
    top = info.get_app_top_frame()
    assert top == info.stack[0]


def test_app_top_root_module():
    info = ErrorInfo(
        name="unknown",
        version="unknown",
        fingerprint=uuid.uuid4(),
        stack=SAMPLE_STACK,
        exception=ValueError(),
        root_module="tests.sample",
    )
    top = info.get_app_top_frame()
    assert top == info.stack[1]



# ---------------------------------------------------------------------
# Gufo Err: BaseFormatter class
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""BaseFormatter class."""

# Python modules
from abc import ABC, abstractmethod
from typing import Iterable

# GufoLabs modules
from ..types import CodePosition, ErrorInfo, ExceptionStub, FrameInfo

DEFAULT_PRIMARY_CHAR = "~"
DEFAULT_SECONDARY_CHAR = "^"


class BaseFormatter(ABC):
    """Abstract base class for formatters.

    Formatters process [ErrorInfo][gufo.err.ErrorInfo]
    instances and produces human-readable output.

    All formatters must implement `iter_format` method.

    Args:
        primary_char: Caret primary char.
        secondary_char: Caret secondary char.
    """

    def __init__(
        self: "BaseFormatter",
        primary_char: str = DEFAULT_PRIMARY_CHAR,
        secondary_char: str = DEFAULT_SECONDARY_CHAR,
    ) -> None:
        self.primary_char = primary_char
        self.secondary_char = secondary_char

    def format(self: "BaseFormatter", err: ErrorInfo) -> str:
        """Format ErrorInfo to human-readable string.

        Args:
            err: ErrorInfo instance

        Returns:
            Human-readable output.
        """
        return "\n".join(self.iter_format(err))

    @abstractmethod
    def iter_format(self: "BaseFormatter", err: ErrorInfo) -> Iterable[str]:
        """Iterator yielding human-redable lines.

        Process ErrorInfo instance and yield humar-readable
        lines one-by-one.

        Args:
            err: ErrorInfo instance

        Returns:
            Iterator yieldig formatted lines.
        """

    def traceback_message(self: "BaseFormatter") -> str:
        """Get proper traceback message.

        Returns:
            String like "Traceback (most resent call last):"
        """
        return "Traceback (most resent call last):"

    def iter_stack(
        self: "BaseFormatter", err: ErrorInfo
    ) -> Iterable[FrameInfo]:
        """Iterate stack according to direction.

        Args:
            err: ErrorInfo instance.

        Returns:
            Iterable of FrameInfo
        """
        yield from err.stack

    def get_caret(
        self: "BaseFormatter",
        line: str,
        pos: CodePosition,
        indent: int,
        dedent: int = 0,
    ) -> str:
        """Generate caret for code position.

        Carret has a format:
        ```
        <spaces><primary chars...><secondary chars...><primary chars...>
        ```.

        Args:
            line: Current unstripped line of code
            pos: CodePositio
            indent: Add `indent` leading spaces
            dedent: Remove `indent` leading spaces
        """
        if pos.start_line != pos.end_line:
            msg = "Position must be on single line"
            raise ValueError(msg)
        # Leading spaces
        leading = " " * (pos.start_col + indent - dedent)
        # Parse AST and find anchors
        anchor = pos.anchor
        if not anchor:
            # Fill everything with secondary char
            carret_len = pos.end_col - pos.start_col
            return f"{leading}{self.secondary_char * carret_len}"
        # <primary...><secondary...><primary...>
        prolog = self.primary_char * (anchor.left - pos.start_col)
        middle = self.secondary_char * (anchor.right - anchor.left)
        epilog = self.primary_char * (pos.end_col - anchor.right)
        return f"{leading}{prolog}{middle}{epilog}"

    @staticmethod
    def get_exception_summary(x: BaseException) -> str:
        """Format exception to summary string.

        Args:
            x: Exception instance

        Returns:
            Formatted string
        """
        # Class name
        if isinstance(x, ExceptionStub):
            cls_name = x.kls
            info = str(x.args[0]) if x.args else ""
        else:
            cls_name = x.__class__.__name__
            info = str(x)
        # Format result
        if info:
            return f"{cls_name}: {info}"
        return cls_name



# ---------------------------------------------------------------------
# Gufo Err: iter_frames tests
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import os
from typing import Optional

# Gufo Labs modules
from gufo.err import (
    HAS_CODE_POSITION,
    Anchor,
    CodePosition,
    FrameInfo,
    SourceInfo,
    exc_traceback,
    iter_frames,
)
from tests.sample.trace import entry

cwd = os.getcwd()


def MaybeCodePosition(
    *,
    start_line: int,
    end_line: int,
    start_col: int,
    end_col: int,
    anchor_left: Optional[int] = None,
    anchor_right: Optional[int] = None,
) -> Optional[CodePosition]:
    if HAS_CODE_POSITION:
        if anchor_left is None or anchor_right is None:
            anchor = None
        else:
            anchor = Anchor(left=anchor_left, right=anchor_right)
        return CodePosition(
            start_line=start_line,
            end_line=end_line,
            start_col=start_col,
            end_col=end_col,
            anchor=anchor,
        )
    return None


def to_full_path(*args) -> str:
    """Convert relative path to full path."""
    return os.path.join(cwd, *args)


SAMPLE_FRAMES = [
    FrameInfo(
        name="entry",
        source=SourceInfo(
            file_name=to_full_path("tests", "sample", "trace.py"),
            first_line=8,
            current_line=15,
            lines=[
                "    x += 1",
                "    oops()",
                "",
                "",
                "def entry():",
                "    s = 2",
                "    s += 1",
                "    to_oops()",
            ],
            pos=MaybeCodePosition(
                start_line=15,
                end_line=15,
                start_col=4,
                end_col=13,
            ),
        ),
        locals={"s": 3},
        module="tests.sample.trace",
    ),
    FrameInfo(
        name="to_oops",
        source=SourceInfo(
            file_name=to_full_path("tests", "sample", "trace.py"),
            first_line=2,
            current_line=9,
            lines=[
                '    msg = "oops"',
                "    raise RuntimeError(msg)",
                "",
                "",
                "def to_oops():",
                "    x = 1",
                "    x += 1",
                "    oops()",
                "",
                "",
                "def entry():",
                "    s = 2",
                "    s += 1",
                "    to_oops()",
            ],
            pos=MaybeCodePosition(
                start_line=9, end_line=9, start_col=4, end_col=10
            ),
        ),
        locals={"x": 2},
        module="tests.sample.trace",
    ),
    FrameInfo(
        name="oops",
        source=SourceInfo(
            file_name=to_full_path("tests", "sample", "trace.py"),
            first_line=1,
            current_line=3,
            lines=[
                "def oops():",
                '    msg = "oops"',
                "    raise RuntimeError(msg)",
                "",
                "",
                "def to_oops():",
                "    x = 1",
                "    x += 1",
                "    oops()",
                "",
            ],
            pos=MaybeCodePosition(
                start_line=3, end_line=3, start_col=4, end_col=27
            ),
        ),
        locals={"msg": "oops"},
        module="tests.sample.trace",
    ),
]


def test_iter_frames():
    """Call the function which raises an exception."""
    try:
        entry()
        msg = "No trace"
        raise AssertionError(msg)
    except RuntimeError:
        # First frame is from this module and may vary
        # on code changes and pytest version bumping.
        # So we discard it.
        frames = list(iter_frames(exc_traceback()))[1:]
        assert frames == SAMPLE_FRAMES



# ---------------------------------------------------------------------
# Gufo Err: failfast module
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""Fail-fast handlers.

Fail-fast handlers are launched on every unnhandled
exception and have to deside if the error is fatal
and the process must be terminated quickly.
See [gufo.err.abs.failfast.BaseFailFast] for details.

The following fail-fast handlers are available out-of-the-box:

* [AlwaysFailFast][gufo.err.failfast.always.AlwaysFailFast]:
  Triggers fail-fast unconditionaly.
* [NeverFailFast][gufo.err.failfast.never.NeverFailFast]:
  Never triggers fail-fast.
* [TypeMatchFailFast][gufo.err.failfast.typematch.TypeMatchFailFast]:
  Fail-fast on the given exception types and optional substrings.
* [TypesFailFast][gufo.err.failfast.types.TypesFailFast]:
  Fail-fast on the given list of exception types.
"""



# ---------------------------------------------------------------------
# Gufo Err: Serialize/deserialize
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""ErrInfo serialization/deserialization primitives."""

# Python modules
import datetime
import json
import uuid
from typing import Any, Dict, Union

# Gufo Labs modules
from .types import ErrorInfo, ExceptionStub, FrameInfo, SourceInfo

CODEC_TYPE = "errorinfo"
CURRENT_VERSION = "1.0"


def __q_x_class(e: BaseException) -> str:
    """Get exception class.

    Args:
        e: Exception instance

    Returns:
        Serialized exception class name
    """
    mod = e.__class__.__module__
    ncls = e.__class__.__name__
    if mod == "builtins":
        return ncls
    return f"{mod}.{ncls}"


def __q_var(x: Any) -> Union[str, int, float]:  # noqa: ANN401
    """Convert variable to the JSON-encodable form.

    Args:
        x: Exception argument

    Returns:
        JSON-serializeable form of argument
    """
    if isinstance(x, (int, float, str)):
        return x
    return str(x)


def __q_frame_info(fi: FrameInfo) -> Dict[str, Any]:
    """Convert FrameInfo into JSON-serializeable form.

    Args:
        fi: FrameInfo instance

    Returns:
        Serialized dict
    """
    r = {
        "name": fi.name,
        "module": fi.module,
        "locals": {x: __q_var(y) for x, y in fi.locals.items()},
    }
    if fi.source:
        r["source"] = __q_source(fi.source)
    return r


def __q_source(si: SourceInfo) -> Dict[str, Any]:
    """Convert SourceInfo into JSON-serializeable form.

    Args:
        si: SourceInfo instance

    Returns:
        Serialized dict
    """
    return {
        "file_name": si.file_name,
        "first_line": si.first_line,
        "current_line": si.current_line,
        "lines": si.lines,
    }


def __q_exception(e: BaseException) -> Dict[str, Any]:
    """Convery exception into JSON-serializeable form.

    Args:
        e: BaseException instance

    Returns:
        Serialized dict
    """
    return {
        "class": __q_x_class(e),
        "args": [__q_var(x) for x in e.args],
    }


def to_dict(info: ErrorInfo) -> Dict[str, Any]:
    """Serialize ErrorInfo to a dict of primitive types.

    Args:
        info: ErrorInfo instance.

    Returns:
        Dict of primitive types (str, int, float).
    """
    r = {
        "$type": CODEC_TYPE,
        "$version": CURRENT_VERSION,
        "name": info.name,
        "version": info.version,
        "fingerprint": str(info.fingerprint),
        "exception": __q_exception(info.exception),
        "stack": [__q_frame_info(x) for x in info.stack],
    }
    if info.timestamp:
        r["timestamp"] = info.timestamp.isoformat()
    if info.root_module:
        r["root_module"] = info.root_module
    return r


def to_json(info: ErrorInfo) -> str:
    """Serialize ErrorInfo to JSON string.

    Args:
        info: ErrorInfo instance.

    Returns:
        json-encoded string.
    """
    return json.dumps(to_dict(info))


def from_dict(data: Dict[str, Any]) -> ErrorInfo:
    """Deserealize Dict to ErrorInfo.

    Args:
        data: Result of to_dict

    Returns:
        ErrorInfo instance

    Raises:
        ValueError: if required key is missed.
    """

    def get(d: Dict[str, Any], name: str) -> Any:  # noqa: ANN401
        """Get the key's value from the dictionary.

        Args:
            d: Data dictionary
            name: Key name.

        Returns:
            Value

        Raises:
            ValueError if key is missed.
        """
        x = d.get(name)
        if x is None:
            msg = f"{name} is required"
            raise ValueError(msg)
        return x

    def get_fi(d: Dict[str, Any]) -> FrameInfo:
        source = get_si(d["source"]) if d.get("source") else None
        return FrameInfo(
            name=get(d, "name"),
            module=get(d, "module"),
            locals=get(d, "locals"),
            source=source,
        )

    def get_si(d: Dict[str, Any]) -> SourceInfo:
        return SourceInfo(
            file_name=get(d, "file_name"),
            first_line=get(d, "first_line"),
            current_line=get(d, "current_line"),
            lines=get(d, "lines"),
        )

    # Check incoming data is dict
    if not isinstance(data, dict):
        msg = "dict required"
        raise ValueError(msg)
    # Check data has proper type signature
    ci_type = get(data, "$type")
    if ci_type != CODEC_TYPE:
        msg = "Invalid $type"
        raise ValueError(msg)
    # Check version
    ci_version = get(data, "$version")
    if ci_version != CURRENT_VERSION:
        msg = "Unknown $version"
        raise ValueError(msg)
    # Process timestamp
    src_ts = data.get("timestamp")
    ts = datetime.datetime.fromisoformat(src_ts) if src_ts else None
    # Exception
    exc = get(data, "exception")
    # Stack
    stack = [get_fi(x) for x in get(data, "stack")]
    # Set exception stub
    return ErrorInfo(
        name=get(data, "name"),
        version=get(data, "version"),
        fingerprint=uuid.UUID(get(data, "fingerprint")),
        timestamp=ts,
        stack=stack,
        exception=ExceptionStub(kls=exc["class"], args=exc["args"]),
        root_module=data.get("root_module"),
    )


def from_json(data: str) -> ErrorInfo:
    """Deserialize ErrorInfo from JSON string.

    Args:
        data: JSON string

    Returns:
        ErrorInfo instance

    Raises:
        ValueError: if required key is missed.
    """
    return from_dict(json.loads(data))



# ---------------------------------------------------------------------
# Gufo Err
# ---------------------------------------------------------------------
# Copyright (C) 2022-25, Gufo Labs
# ---------------------------------------------------------------------
"""Human-readable error reporting.

Attributes:
    __version__: Current version.
    HAS_CODE_POSITION: True, if Python interpreter supports
        exact code positions  (Python 3.11+)
"""

# Gufo Labs modules
from .abc.failfast import BaseFailFast
from .abc.middleware import BaseMiddleware
from .err import Err, err
from .frame import HAS_CODE_POSITION, exc_traceback, iter_frames
from .logger import logger
from .types import (
    Anchor,
    CodePosition,
    ErrorInfo,
    FrameInfo,
    SourceInfo,
)

__version__: str = "0.5.0"



# ---------------------------------------------------------------------
# Gufo Err: test formatters
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import datetime
import os
import uuid

# Third-party modules
import pytest

# Gufo Err modules
from gufo.err import CodePosition, ErrorInfo, FrameInfo, SourceInfo
from gufo.err.formatter.extend import ExtendFormatter
from gufo.err.formatter.loader import get_formatter
from gufo.err.formatter.terse import TerseFormatter

SAMPLE_ERR = ErrorInfo(
    name="my-test",
    version="3.14",
    fingerprint=uuid.UUID("be8ccd86-3661-434c-8569-40dd65d9860a"),
    timestamp=datetime.datetime(2023, 9, 1, 6, 31, 16, 79222),
    exception=RuntimeError("oops"),
    stack=[
        FrameInfo(
            name="test_iter_frames",
            source=SourceInfo(
                file_name=os.path.join(
                    os.sep, "app", "tests", "test_frames.py"
                ),
                first_line=167,
                current_line=174,
                lines=[
                    "    ),",
                    "]",
                    "",
                    "",
                    "def test_iter_frames():",
                    '    """Call the function which raises an exception."""',
                    "    try:",
                    "        entry()",
                    '        msg = "No trace"',
                    "        raise AssertionError(msg)",
                    "    except RuntimeError:",
                    "        frames = list(iter_frames(exc_traceback()))",
                    "        assert frames == SAMPLE_FRAMES",
                ],
                pos=CodePosition(
                    start_line=174,
                    end_line=174,
                    start_col=8,
                    end_col=15,
                    anchor=None,
                ),
            ),
            locals={},
            module="tests.test_frames",
        ),
        FrameInfo(
            name="entry",
            source=SourceInfo(
                file_name=os.path.join(
                    os.sep, "app", "tests", "sample", "trace.py"
                ),
                first_line=8,
                current_line=15,
                lines=[
                    "    x += 1",
                    "    oops()",
                    "",
                    "",
                    "def entry():",
                    "    s = 2",
                    "    s += 1",
                    "    to_oops()",
                ],
                pos=CodePosition(
                    start_line=15,
                    end_line=15,
                    start_col=4,
                    end_col=13,
                    anchor=None,
                ),
            ),
            locals={"s": 3},
            module="tests.sample.trace",
        ),
        FrameInfo(
            name="to_oops",
            source=SourceInfo(
                file_name=os.path.join(
                    os.sep, "app", "tests", "sample", "trace.py"
                ),
                first_line=2,
                current_line=9,
                lines=[
                    '    msg = "oops"',
                    "    raise RuntimeError(msg)",
                    "",
                    "",
                    "def to_oops():",
                    "    x = 1",
                    "    x += 1",
                    "    oops()",
                    "",
                    "",
                    "def entry():",
                    "    s = 2",
                    "    s += 1",
                    "    to_oops()",
                ],
                pos=CodePosition(
                    start_line=9,
                    end_line=9,
                    start_col=4,
                    end_col=10,
                    anchor=None,
                ),
            ),
            locals={"x": 2},
            module="tests.sample.trace",
        ),
        FrameInfo(
            name="oops",
            source=SourceInfo(
                file_name=os.path.join(
                    os.sep, "app", "tests", "sample", "trace.py"
                ),
                first_line=1,
                current_line=3,
                lines=[
                    "def oops():",
                    '    msg = "oops"',
                    "    raise RuntimeError(msg)",
                    "",
                    "",
                    "def to_oops():",
                    "    x = 1",
                    "    x += 1",
                    "    oops()",
                    "",
                ],
                pos=CodePosition(
                    start_line=3,
                    end_line=3,
                    start_col=4,
                    end_col=27,
                    anchor=None,
                ),
            ),
            locals={"msg": "oops"},
            module="tests.sample.trace",
        ),
    ],
)

TERSE_RESULT = """Error: be8ccd86-3661-434c-8569-40dd65d9860a
Traceback (most resent call last):
  File "/app/tests/test_frames.py", line 174, in test_iter_frames
    entry()
    ^^^^^^^
  File "/app/tests/sample/trace.py", line 15, in entry
    to_oops()
    ^^^^^^^^^
  File "/app/tests/sample/trace.py", line 9, in to_oops
    oops()
    ^^^^^^
  File "/app/tests/sample/trace.py", line 3, in oops
    raise RuntimeError(msg)
    ^^^^^^^^^^^^^^^^^^^^^^^
RuntimeError: oops"""

EXTEND_RESULT = "\n".join(
    [
        "Error: be8ccd86-3661-434c-8569-40dd65d9860a",
        "RuntimeError: oops",
        "Traceback (most resent call last):",
        "-------------------------------------------------------------------------------",
        "File: /app/tests/test_frames.py (line 174)",
        "  167         ),",
        "  168     ]",
        "  169     ",
        "  170     ",
        "  171     def test_iter_frames():",
        '  172         """Call the function which raises an exception."""',
        "  173         try:",
        "  174 ==>         entry()",
        "                  ^^^^^^^",
        '  175             msg = "No trace"',
        "  176             raise AssertionError(msg)",
        "  177         except RuntimeError:",
        "  178             frames = list(iter_frames(exc_traceback()))",
        "  179             assert frames == SAMPLE_FRAMES",
        "-------------------------------------------------------------------------------",
        "File: /app/tests/sample/trace.py (line 15)",
        "    8         x += 1",
        "    9         oops()",
        "   10     ",
        "   11     ",
        "   12     def entry():",
        "   13         s = 2",
        "   14         s += 1",
        "   15 ==>     to_oops()",
        "              ^^^^^^^^^",
        "Locals:",
        "                   s = 3",
        "-------------------------------------------------------------------------------",
        "File: /app/tests/sample/trace.py (line 9)",
        '    2         msg = "oops"',
        "    3         raise RuntimeError(msg)",
        "    4     ",
        "    5     ",
        "    6     def to_oops():",
        "    7         x = 1",
        "    8         x += 1",
        "    9 ==>     oops()",
        "              ^^^^^^",
        "   10     ",
        "   11     ",
        "   12     def entry():",
        "   13         s = 2",
        "   14         s += 1",
        "   15         to_oops()",
        "Locals:",
        "                   x = 2",
        "-------------------------------------------------------------------------------",
        "File: /app/tests/sample/trace.py (line 3)",
        "    1     def oops():",
        '    2         msg = "oops"',
        "    3 ==>     raise RuntimeError(msg)",
        "              ^^^^^^^^^^^^^^^^^^^^^^^",
        "    4     ",
        "    5     ",
        "    6     def to_oops():",
        "    7         x = 1",
        "    8         x += 1",
        "    9         oops()",
        "   10     ",
        "Locals:",
        "                 msg = 'oops'",
        "-------------------------------------------------------------------------------",
    ]
)


def test_get_caret_error() -> None:
    pos = CodePosition(
        start_line=10, end_line=11, start_col=4, end_col=4, anchor=None
    )
    formatter = TerseFormatter()
    with pytest.raises(ValueError):
        formatter.get_caret("    test()", pos=pos, indent=0)


def test_loader_invalid() -> None:
    with pytest.raises(ValueError):
        get_formatter("invalid")


def test_loader_terse() -> None:
    formatter = get_formatter("terse")
    assert isinstance(formatter, TerseFormatter)


def test_loader_extend() -> None:
    formatter = get_formatter("extend")
    assert isinstance(formatter, ExtendFormatter)


def test_terse() -> None:
    formatter = get_formatter("terse")
    r = formatter.format(SAMPLE_ERR)
    assert r == TERSE_RESULT


def test_extend() -> None:
    formatter = get_formatter("extend")
    r = formatter.format(SAMPLE_ERR)
    assert r == EXTEND_RESULT



# ---------------------------------------------------------------------
# Gufo Err: format module
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""ErrInfo formatters.

Formatters process [ErrorInfo][gufo.err.ErrorInfo]
structure and produces human-readable output.

Available out-of-box:

* [TerseFormatter][gufo.err.formatter.terse.TerseFormatter]:
  condensed minimal output.
* [ExtendFormatter][gufo.err.formatter.extend.ExtendFormatter]:
  extended detailed output.

Configured formatter instances can be obtained via
[get_formatter()][gufo.err.formatter.loader.get_formatter]
function.
"""



# ---------------------------------------------------------------------
# Gufo Err: test ErrorInfoMiddleware
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Third-party modules
import re

import pytest

# Gufo Labs modules
from gufo.err import Err
from gufo.err.codec import from_json
from gufo.err.compressor import Compressor

from .util import log_capture

rx_log_path = re.compile(r"Writing error info into (\S+)")


def test_unwritable_path():
    with pytest.raises(ValueError):
        Err().setup(error_info_path="/a/b/c/d")


def test_invalid_compress(tmpdir):
    with pytest.raises(ValueError):
        Err().setup(
            error_info_path=tmpdir.mkdir("errinfo"), error_info_compress="rar"
        )


@pytest.mark.parametrize("compress", [None, "gz", "bz2", "xz"])
def test_compress(tmpdir, compress):
    err_info_path = tmpdir.mkdir("errinfo")
    # Setup
    err = Err().setup(
        format=None,
        error_info_path=err_info_path,
        error_info_compress=compress,
    )
    # Generate error
    try:
        msg = "oops"
        raise RuntimeError(msg)
    except RuntimeError:
        with log_capture() as buffer:
            err.process()
            output = buffer.getvalue()
    # Check error info
    match = rx_log_path.search(output)
    assert match
    ei_file = match.group(1)
    if compress is None:
        assert ei_file.endswith(".json")
    else:
        assert ei_file.endswith(f".json.{compress}")
    # Check crashinfo is written
    assert len(err_info_path.listdir()) == 1
    ei_path = err_info_path.listdir()[0]
    # Compare to logged
    assert ei_path == ei_file
    # Check compressor suffix
    compressor = Compressor(format=compress)
    assert ei_file.endswith(compressor.suffix)
    # Try to read file
    with open(ei_path, "rb") as f:
        data = compressor.decode(f.read()).decode("utf-8")
    ei = from_json(data)
    # Check error info
    fn = f"{ei.fingerprint}.json{compressor.suffix}"
    assert ei_file.endswith(fn)


@pytest.mark.parametrize("compress", [None, "gz", "bz2", "xz"])
def test_seen(tmpdir, compress):
    err_info_path = tmpdir.mkdir("errinfo")
    # Setup
    err = Err().setup(
        format=None,
        error_info_path=err_info_path,
        error_info_compress=compress,
    )
    # Generate error
    try:
        msg = "oops"
        raise RuntimeError(msg)
    except RuntimeError:
        with log_capture() as buffer:
            err.process()
            err.process()  # Must seen the error
            output = buffer.getvalue()
    # Check error info
    assert "is already registered" in output



# ---------------------------------------------------------------------
# Gufo Err: BaseResponse class
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""BaseMiddleware."""

# Python modules
from abc import ABC, abstractmethod

# GufoLabs modules
from ..types import ErrorInfo


class BaseMiddleware(ABC):
    """Abstract base type for error processing middleware.

    Middleware must implement `process` method.
    """

    @abstractmethod
    def process(self: "BaseMiddleware", info: ErrorInfo) -> None:
        """Process the error.

        Args:
            info: ErrorInfo instance with detailed error information.
        """
        ...



# ---------------------------------------------------------------------
# Gufo Err: test TracebackResponse
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import os
import re
from typing import Any, Dict

# Third-party modules
import pytest

# Gufo Labs modules
from gufo.err import Err

from .sample.trace import entry
from .util import log_capture

RESAMPLE = False  # Change to True to rewrite sample files

rx_file = re.compile(r"^(\s*File:?\s*\"?)(.*/gufo_err)(/.*)$", re.MULTILINE)
rx_repr = re.compile("(= <.*? at 0x)([0-9a-f]+?)(>)", re.MULTILINE)
rx_caret = re.compile(r"\n +~*\^+~*$", re.MULTILINE)


def clean_config(src: str) -> str:
    src = rx_file.sub(r"\1\3", src)
    src = rx_repr.sub(r"\1aaaa\3", src)
    # Remove python 3.11+ caret
    return rx_caret.sub("", src)


def get_sample_path(fmt: str) -> str:
    return os.path.join("tests", "data", "trace", f"{fmt}.txt")


def get_sample(fmt: str) -> str:
    with open(get_sample_path(fmt)) as f:
        return f.read()


def set_sample(fmt: str, sample: str) -> None:
    with open(get_sample_path(fmt), "w") as f:
        f.write(sample)


def test_ci_resample() -> None:
    if RESAMPLE:
        assert "CI" not in os.environ, "RESAMPLE must not be set for CI"


@pytest.mark.parametrize(
    "cfg",
    [
        {},
        {"format": "terse"},
        {"format": "extend"},
    ],
)
def test_format(cfg: Dict[str, Any]) -> None:
    fmt = cfg.get("format", "terse")
    err = Err().setup(**cfg)
    cfg = {}  # Reset local vars
    try:
        entry()
    except Exception:
        with log_capture() as buffer:
            err.process()
            output = buffer.getvalue()
    if RESAMPLE:
        set_sample(fmt, output)
    assert clean_config(output) == clean_config(get_sample(fmt))



# ---------------------------------------------------------------------
# Gufo Err: get_formatter function
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""get_formatter implementation."""

# Gufo Err modules
from ..abc.formatter import (
    DEFAULT_PRIMARY_CHAR,
    DEFAULT_SECONDARY_CHAR,
    BaseFormatter,
)


def get_formatter(
    format: str = "terse",
    *,
    primary_char: str = DEFAULT_PRIMARY_CHAR,
    secondary_char: str = DEFAULT_SECONDARY_CHAR,
) -> BaseFormatter:
    """Configure and return formatter instance.

    Args:
        format: Formatter name, one of:
            * `terse`
            * `extend`
        primary_char: Caret primary char.
        secondary_char: Caret secondary char.

    Returns:
        Formatter instance.
    """
    if format == "terse":
        from .terse import TerseFormatter

        return TerseFormatter(
            primary_char=primary_char, secondary_char=secondary_char
        )
    if format == "extend":
        from .extend import ExtendFormatter

        return ExtendFormatter(
            primary_char=primary_char, secondary_char=secondary_char
        )
    msg = f"Invalid formatter: {format}"
    raise ValueError(msg)



# ---------------------------------------------------------------------
# Gufo Labs: CLI test
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# See LICENSE.md for details
# ---------------------------------------------------------------------

# Python modules
import os
import tempfile
from collections import defaultdict

# Third-party modules
import pytest

# Gufo Err modules
from gufo.err import Err, __version__
from gufo.err.cli import Cli, ExitCode

FAKE_UUID = "e5d2e3e0-d6b9-415f-ba09-5eadbe4b38ad"


def fn1():
    raise ValueError


def fn2():
    msg = "foobar"
    raise NameError(msg)


def fn3():
    raise TypeError


def fn4():
    raise NotImplementedError


@pytest.fixture(scope="module")
def crashinfo():
    """Create crashinfo directory and populate it with several crashes."""
    with tempfile.TemporaryDirectory() as tmpdir:
        for codec, fn in [
            (None, fn1),
            ("gz", fn2),
            ("bz2", fn3),
            ("xz", fn4),
        ]:
            err = Err().setup(
                name=f"fmt-{codec}",
                format=None,
                error_info_path=tmpdir,
                error_info_compress=codec,
            )
            try:
                fn()
            except Exception:
                err.process()
        yield tmpdir


def test_help_short() -> None:
    with pytest.raises(SystemExit):
        Cli().run(["-h"])


def test_help_long() -> None:
    with pytest.raises(SystemExit):
        Cli().run(["--help"])


def test_version(capsys) -> None:
    r = Cli().run(["version"])
    assert r == ExitCode.OK
    out = capsys.readouterr().out
    assert out == f"Gufo Err {__version__}\n"


def test_crashinfo_setup(crashinfo) -> None:
    assert os.path.exists(crashinfo)
    lst = os.listdir(crashinfo)
    # Check length
    assert len(lst) == 4
    # Check suffixes
    r = defaultdict(int)
    for fn in lst:
        r[fn.split(".", 1)[1]] += 1
    assert r.get("json") == 1
    assert r.get("json.gz") == 1
    assert r.get("json.bz2") == 1
    assert r.get("json.xz") == 1


def test_list_invalid(capsys) -> None:
    path = "/nonexistend/directory"
    r = Cli().run(["-p", path, "list"])
    assert r == ExitCode.NOT_EXISTS
    out = capsys.readouterr().out
    assert out == f"Error: {path} is not exists\n"


def test_list(capsys, crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "list"])
    assert r == ExitCode.OK
    out = capsys.readouterr().out
    assert "TypeError" in out
    assert "ValueError" in out
    assert "NameError: foobar" in out
    assert "NotImplementedError" in out
    assert "fmt-None" in out
    assert "fmt-gz" in out
    assert "fmt-bz2" in out
    assert "fmt-xz" in out
    assert "test_cli.py:57" in out
    # assert out == ""


def test_list_all(capsys, crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "list", "all"])
    assert r == ExitCode.OK
    out = capsys.readouterr().out
    assert "TypeError" in out
    assert "ValueError" in out
    assert "NameError: foobar" in out
    assert "NotImplementedError" in out
    assert "fmt-None" in out
    assert "fmt-gz" in out
    assert "fmt-bz2" in out
    assert "fmt-xz" in out
    assert "test_cli.py:57" in out
    # assert out == ""


def test_list_two(capsys, crashinfo) -> None:
    all_fp = [fn.split(".")[0] for fn in os.listdir(crashinfo)]
    include = all_fp[:2]
    exclude = all_fp[2:]
    r = Cli().run(["-p", crashinfo, "list", *include])
    assert r == ExitCode.OK
    out = capsys.readouterr().out
    for fp in include:
        assert fp in out
    for fp in exclude:
        assert fp not in out


def test_list_syntax_error(capsys, crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "list", "invalid"])
    assert r == ExitCode.SYNTAX


def test_view_invalid(capsys, crashinfo) -> None:
    path = "/nonexistend/directory"
    r = Cli().run(["-p", path, "view", FAKE_UUID])
    assert r == ExitCode.NOT_EXISTS
    out = capsys.readouterr().out
    assert out == f"Error: {path} is not exists\n"


def test_view_empty(crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "view"])
    assert r == ExitCode.OK


def test_view_cannot_read(crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "view", FAKE_UUID])
    assert r == ExitCode.CANNOT_READ


def test_view_invalid_args(crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "view", "-f", "custom", FAKE_UUID])
    assert r == ExitCode.INVALID_ARGS


def test_view(capsys, crashinfo) -> None:
    all_fp = [fn.split(".")[0] for fn in os.listdir(crashinfo)]
    r = Cli().run(["-p", crashinfo, "view", *all_fp])
    assert r == ExitCode.OK


def test_view_terse(capsys, crashinfo) -> None:
    all_fp = [fn.split(".")[0] for fn in os.listdir(crashinfo)]
    r = Cli().run(["-p", crashinfo, "view", "-f", "terse", *all_fp])
    assert r == ExitCode.OK
    out = capsys.readouterr().out
    assert "TypeError" in out
    assert "ValueError" in out
    assert "NameError: foobar" in out
    assert "NotImplementedError" in out
    for fp in all_fp:
        assert fp in out


def test_view_extend(capsys, crashinfo) -> None:
    all_fp = [fn.split(".")[0] for fn in os.listdir(crashinfo)]
    r = Cli().run(["-p", crashinfo, "view", "--format", "extend", *all_fp])
    assert r == ExitCode.OK
    out = capsys.readouterr().out
    assert "TypeError" in out
    assert "ValueError" in out
    assert "NameError: foobar" in out
    assert "NotImplementedError" in out
    for fp in all_fp:
        assert fp in out


def test_view_all(capsys, crashinfo):
    all_fp = [fn.split(".")[0] for fn in os.listdir(crashinfo)]
    r = Cli().run(["-p", crashinfo, "view", "all"])
    assert r == ExitCode.OK
    out = capsys.readouterr().out
    for fp in all_fp:
        assert fp in out


def test_view_all_star(capsys, crashinfo):
    all_fp = [fn.split(".")[0] for fn in os.listdir(crashinfo)]
    r = Cli().run(["-p", crashinfo, "view", "*"])
    assert r == ExitCode.OK
    out = capsys.readouterr().out
    for fp in all_fp:
        assert fp in out


def test_view_two(capsys, crashinfo) -> None:
    all_fp = [fn.split(".")[0] for fn in os.listdir(crashinfo)]
    include = all_fp[:2]
    exclude = all_fp[2:]
    r = Cli().run(["-p", crashinfo, "view", *include])
    assert r == ExitCode.OK
    out = capsys.readouterr().out
    for fp in include:
        assert fp in out
    for fp in exclude:
        assert fp not in out


def test_view_syntax_error(capsys, crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "view", "invalid"])
    assert r == ExitCode.SYNTAX


@pytest.mark.parametrize(
    ("t", "w", "exp"),
    [
        ("12345", 6, "12345 "),
        ("12345", 5, "12345"),
        ("123456", 5, "12..."),
    ],
)
def test_col(t: str, w: int, exp: str) -> None:
    r = Cli.col(t, w)
    assert r == exp


@pytest.mark.parametrize(
    ("t", "w", "exp"),
    [
        ("12345", 6, "12345 "),
        ("12345", 5, "12345"),
        ("123456", 5, "...56"),
    ],
)
def test_rcol(t: str, w: int, exp: str) -> None:
    r = Cli.rcol(t, w)
    assert r == exp


def test_clear_empty(crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "clear"])
    assert r == ExitCode.OK


def test_clear_syntax_error(capsys, crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "clear", "invalid"])
    assert r == ExitCode.SYNTAX


def test_clear_cannot_read(crashinfo) -> None:
    r = Cli().run(["-p", crashinfo, "clear", FAKE_UUID])
    assert r == ExitCode.CANNOT_READ


def test_clear_invalid(capsys, crashinfo) -> None:
    path = "/nonexistend/directory"
    r = Cli().run(["-p", path, "clear", FAKE_UUID])
    assert r == ExitCode.NOT_EXISTS
    out = capsys.readouterr().out
    assert out == f"Error: {path} is not exists\n"


# Keep this test as latest in the modules
def test_clear(crashinfo) -> None:
    def ls():
        return [fn.split(".")[0] for fn in os.listdir(crashinfo)]

    # Delete one
    all_fp = ls()
    assert len(all_fp) == 4
    to_delete = all_fp[0]
    r = Cli().run(["-p", crashinfo, "clear", to_delete])
    assert r == ExitCode.OK
    all_fp = ls()
    assert to_delete not in all_fp
    assert len(all_fp) == 3
    # Delete two
    to_delete = all_fp[:2]
    r = Cli().run(["-p", crashinfo, "clear", *to_delete])
    assert r == ExitCode.OK
    all_fp = ls()
    for fp in to_delete:
        assert fp not in all_fp
    assert len(all_fp) == 1
    # Delete all
    r = Cli().run(["-p", crashinfo, "clear", "all"])
    assert r == ExitCode.OK
    assert not ls()



# ---------------------------------------------------------------------
# Gufo Err: test Compressor
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Third-party modules
import pytest

# Gufo Labs Modules
from gufo.err.compressor import Compressor


def test_invalid_format():
    with pytest.raises(ValueError):
        Compressor(format="rar")


@pytest.mark.parametrize(
    ("path", "expected"),
    [
        ("/a/b/c/xxx.json", None),
        ("/a/b/c/xxx.json.gz", "gz"),
        ("/a/b/c/xxx.json.bz2", "bz2"),
        ("/a/b/c/xxx.json.xz", "xz"),
        ("/a/b/c/xxx.json.rar", None),
    ],
)
def test_get_format(path: str, expected: str) -> None:
    assert Compressor.get_format(path) == expected


@pytest.mark.parametrize(
    ("path", "expected"),
    [
        ("/a/b/c/xxx.json", None),
        ("/a/b/c/xxx.json.gz", "gz"),
        ("/a/b/c/xxx.json.bz2", "bz2"),
        ("/a/b/c/xxx.json.xz", "xz"),
        ("/a/b/c/xxx.json.rar", None),
    ],
)
def test_autodetect(path: str, expected: str) -> None:
    c = Compressor.autodetect(path)
    assert isinstance(c, Compressor)
    src = b"12345"
    cdata = c.encode(src)
    assert c.decode(cdata) == src


@pytest.mark.parametrize(
    ("fmt", "data"),
    [
        (None, b"12345"),
        (
            "gz",
            b"12345",
        ),
        (
            "bz2",
            b"12345",
        ),
        (
            "xz",
            b"12345",
        ),
    ],
)
def test_compressor(fmt: str, data: bytes):
    c = Compressor(format=fmt)
    c_data = c.encode(data)
    s_data = c.decode(c_data)
    assert s_data == data



# ---------------------------------------------------------------------
# Gufo Err: Err tests
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import os
import sys
from typing import Iterable, List, Type
from uuid import UUID

# Third-party modules
import pytest

# Gufo Labs modules
from gufo.err import (
    BaseMiddleware,
    Err,
    ErrorInfo,
    FrameInfo,
    SourceInfo,
)
from gufo.err.failfast.always import AlwaysFailFast
from gufo.err.failfast.never import NeverFailFast
from gufo.err.failfast.typematch import TypeMatchFailFast
from gufo.err.failfast.types import TypesFailFast


def test_unitialized():
    def process():
        try:
            msg = "test"
            raise Exception(msg)
        except Exception:
            err.process()

    err = Err()
    with pytest.raises(RuntimeError):
        process()


def test_double_initialized():
    err = Err()
    err.setup()
    with pytest.raises(RuntimeError):
        err.setup()


def test_empty_process():
    err = Err()
    err.setup()
    err.process()


def test_invalid_hash():
    err = Err()
    with pytest.raises(ValueError):
        err.setup(hash="invalidhash")


@pytest.mark.parametrize(
    ("hash", "data", "expected"),
    [
        ("sha1", ["test"], "a94a8fe5-ccb1-5ba6-9c4c-0873d391e987"),
        ("sha1", ["test", "test2"], "7408bb49-1e0d-5dc3-bac8-e503353e7940"),
        ("sha256", ["test"], "9f86d081-884c-5d65-9a2f-eaa0c55ad015"),
        ("sha256", ["test", "test2"], "98316636-bec9-5dbd-a3b4-60a9d08b8b16"),
        ("md5", ["test"], "098f6bcd-4621-5373-8ade-4e832627b4f6"),
        ("md5", ["test", "test2"], "8f2ab979-2f93-569a-9898-855f12414208"),
    ],
)
def test_hash(hash, data, expected):
    class ZeroHashErr(Err):
        def iter_fingerprint_parts(
            self,
            t: Type[BaseException],
            v: BaseException,
            stack: List[FrameInfo],
        ) -> Iterable[str]:
            yield from data

    err = ZeroHashErr()
    err.setup(hash=hash)
    try:
        msg = "test"
        raise RuntimeError(msg)
    except Exception:
        t, v, _ = sys.exc_info()
        assert t
        assert v
        fp = err._Err__fingerprint(t, v, [])
        assert isinstance(fp, UUID)
        assert str(fp) == expected


@pytest.mark.parametrize(
    ("stack", "root_module", "expected"),
    [
        (
            [
                FrameInfo(
                    name="test_fn",
                    module="tests.test",
                    source=SourceInfo(
                        file_name="tests/test.py",
                        current_line=10,
                        first_line=3,
                        lines=[],
                    ),
                    locals={},
                )
            ],
            None,
            [
                "fp_test",
                "1.0.0",
                "RuntimeError",
                "tests.test",
                "test_fn",
                "10",
            ],
        ),
        (
            [
                FrameInfo(
                    name="lib_fn",
                    module="lib.final.test",
                    source=SourceInfo(
                        file_name="lib/final/test.py",
                        current_line=17,
                        first_line=10,
                        lines=[],
                    ),
                    locals={},
                ),
                FrameInfo(
                    name="proxt_fn",
                    module="lib.proxy.test",
                    source=SourceInfo(
                        file_name="lib/proxy/test.py",
                        current_line=20,
                        first_line=13,
                        lines=[],
                    ),
                    locals={},
                ),
                FrameInfo(
                    name="test_fn",
                    source=SourceInfo(
                        file_name="tests/test.py",
                        current_line=10,
                        first_line=3,
                        lines=[],
                    ),
                    locals={},
                ),
            ],
            None,
            [
                "fp_test",
                "1.0.0",
                "RuntimeError",
                "lib.final.test",
                "lib_fn",
                "17",
            ],
        ),
        (
            [
                FrameInfo(
                    name="lib_fn",
                    module="lib.final.test",
                    source=SourceInfo(
                        file_name="lib/final/test.py",
                        current_line=17,
                        first_line=10,
                        lines=[],
                    ),
                    locals={},
                ),
                FrameInfo(
                    name="proxt_fn",
                    module="lib.proxy.test",
                    source=SourceInfo(
                        file_name="lib/proxy/test.py",
                        current_line=20,
                        first_line=13,
                        lines=[],
                    ),
                    locals={},
                ),
                FrameInfo(
                    name="test_fn",
                    module="tests.test",
                    source=SourceInfo(
                        file_name="tests/test.py",
                        current_line=10,
                        first_line=3,
                        lines=[],
                    ),
                    locals={},
                ),
            ],
            "tests.test",
            [
                "fp_test",
                "1.0.0",
                "RuntimeError",
                "lib.final.test",
                "lib_fn",
                "17",
                "tests.test",
                "test_fn",
                "10",
            ],
        ),
    ],
)
def test_iter_fingerprint_parts(stack, root_module, expected):
    err = Err()
    err.setup(name="fp_test", version="1.0.0", root_module=root_module)
    try:
        msg = "test"
        raise RuntimeError(msg)
    except Exception:
        t, v, _ = sys.exc_info()
        assert t
        assert v
        parts = list(err.iter_fingerprint_parts(t, v, stack))
        assert parts == expected


def test_must_die_no_tb():
    err = Err()
    err.setup()
    try:
        msg = "test"
        raise RuntimeError(msg)
    except Exception:
        t, v, _ = sys.exc_info()
        assert t
        assert v
        err._Err__must_die(t, v, None)


@pytest.mark.parametrize(
    ("chain", "expected"),
    [
        ([NeverFailFast()], False),
        ([AlwaysFailFast()], True),
        ([NeverFailFast(), NeverFailFast()], False),
        ([NeverFailFast(), AlwaysFailFast()], True),
        ([TypesFailFast([RuntimeError])], True),
        ([TypesFailFast([ValueError])], False),
        ([TypesFailFast([RuntimeError, ValueError])], True),
        ([TypeMatchFailFast(RuntimeError)], True),
        ([TypeMatchFailFast(ValueError)], False),
        ([TypeMatchFailFast("builtins.RuntimeError")], True),
        ([TypeMatchFailFast(RuntimeError, match="test")], True),
        ([TypeMatchFailFast(RuntimeError, match="est")], True),
        ([TypeMatchFailFast(RuntimeError, match="nest")], False),
        ([TypeMatchFailFast(RuntimeError, msg="stop")], True),
        ([TypeMatchFailFast(RuntimeError, msg="stop: %s")], True),
    ],
)
def test_must_die(chain, expected):
    err = Err()
    err.setup(fail_fast=chain)
    try:
        msg = "test"
        raise RuntimeError(msg)
    except Exception:
        t, v, tb = sys.exc_info()
        assert t
        assert v
        assert tb
        r = err._Err__must_die(t, v, tb)
        assert r is expected


def test_no_catch_all():
    prev_hook = sys.excepthook
    err = Err()
    err.setup()
    assert sys.excepthook == prev_hook


def test_catch_all():
    prev_hook = sys.excepthook
    err = Err()
    err.setup(catch_all=True)
    assert sys.excepthook == err._Err__process
    sys.excepthook = prev_hook


@pytest.mark.parametrize("exc_class", [SystemExit, KeyboardInterrupt])
def test_process_excluded_exc(exc_class: Type[BaseException]):
    err = Err()
    err.setup()
    try:
        try:
            msg = "test"
            raise exc_class(msg)
        except BaseException:
            err.process()
    except exc_class as e:
        assert e.args[0] == "test"  # noqa: PT017


def test_process_no_tb():
    err = Err()
    err.setup()
    try:
        msg = "test"
        raise RuntimeError(msg)
    except Exception:
        t, v, _ = sys.exc_info()
        assert t
        assert v
        err._Err__process(t, v, None)


@pytest.mark.parametrize("code", [1, 2, 3])
def test_failfast_exit(code):
    def _exit(c):
        nonlocal exit_code
        exit_code = c

    err = Err()
    err.setup(fail_fast_code=code, fail_fast=[AlwaysFailFast()])
    exit_code = 0
    prev_exit = os._exit
    os._exit = _exit
    try:
        msg = "test"
        raise RuntimeError(msg)
    except Exception:
        err.process()
    os._exit = prev_exit
    assert exit_code == code


def test_add_fail_fast():
    class NamedFailFast(NeverFailFast):
        def __init__(self, name: str) -> None:
            super().__init__()
            self.name = name

    err = Err()
    err.setup(fail_fast=[NamedFailFast("f1"), NamedFailFast("f2")])
    err.add_fail_fast(NamedFailFast("f3"))
    err.add_fail_fast(NamedFailFast("f4"))
    chain = [f.name for f in err._Err__failfast_chain]
    assert chain == ["f1", "f2", "f3", "f4"]


def test_middlewre():
    class NamedMiddleware(BaseMiddleware):
        def __init__(self, name: str, fail: bool = False) -> None:
            super().__init__()
            self.name = name
            self.fail = fail

        def process(self, info: ErrorInfo) -> None:
            nonlocal passed
            passed[self.name] = True
            if self.fail:
                msg = "test"
                raise ValueError(msg)

    passed = {}

    err = Err()
    err.setup(middleware=[NamedMiddleware("r1"), NamedMiddleware("r2")])
    err.add_middleware(NamedMiddleware("r3", fail=True))
    err.add_middleware(NamedMiddleware("r4"))
    try:
        msg = "test"
        raise RuntimeError(msg)
    except Exception:
        err.process()
    for i in ("r1", "r2", "r3", "r4"):
        assert passed.get(i) is True


def test_failfast_type_setup():
    err = Err()
    with pytest.raises(ValueError):
        err.setup(fail_fast=[1])


def test_failfast_type_add():
    err = Err()
    err.setup()
    with pytest.raises(ValueError):
        err.add_fail_fast(1)


def test_middleware_type_setup():
    err = Err()
    with pytest.raises(ValueError):
        err.setup(middleware=[1])


def test_middleware_type_add():
    err = Err()
    err.setup()
    with pytest.raises(ValueError):
        err.add_middleware(1)



# ---------------------------------------------------------------------
# Gufo Err: TerseFormatter class
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""TerseFormatter class."""

# Python modules
from typing import Iterable

from ..abc.formatter import BaseFormatter

#  Gufo Err modules
from ..types import ErrorInfo


class TerseFormatter(BaseFormatter):
    """Condensed terse output."""

    def iter_format(self: "TerseFormatter", err: ErrorInfo) -> Iterable[str]:
        """Iterator yielding human-redable lines.

        Process ErrorInfo instance and yield humar-readable
        lines one-by-one.

        Args:
            err: ErrorInfo instance

        Returns:
            Iterator yieldig formatted lines.
        """
        yield f"Error: {err.fingerprint}"
        yield self.traceback_message()
        for fi in self.iter_stack(err):
            if fi.source:
                yield (
                    f'  File "{fi.source.file_name}", '
                    f"line {fi.source.current_line}, in {fi.name}"
                )
                if (
                    fi.source.pos
                    and fi.source.pos.start_line == fi.source.pos.end_line
                ):
                    # Exact position, single line
                    pos = fi.source.pos
                    current_line = fi.source.lines[
                        pos.start_line - fi.source.first_line
                    ]
                    line = current_line.lstrip()
                    yield f"    {line}"
                    # Show caret
                    yield self.get_caret(
                        current_line, pos, 4, len(current_line) - len(line)
                    )
                else:
                    # No position, show current line
                    line = fi.source.lines[
                        fi.source.current_line - fi.source.first_line
                    ].lstrip()
                    yield f"    {line}"
            else:
                yield '  File "<stdin>", line ??? in <module>'
        yield self.get_exception_summary(err.exception)



# ---------------------------------------------------------------------
# Gufo Err: ExtendFormatter class
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""ExtendFormatter."""

# Python modules
from pprint import pformat
from typing import Iterable, Tuple

# Gufo Err modules
from ..abc.formatter import BaseFormatter
from ..types import ErrorInfo, FrameInfo


class ExtendFormatter(BaseFormatter):
    """Extended output.

    Produces extended output with code surroundings
    and variable values.
    """

    SEP = "-" * 79
    MAX_VAR_LEN = 72

    def iter_format(self: "ExtendFormatter", err: ErrorInfo) -> Iterable[str]:
        """Iterator yielding human-redable lines.

        Process ErrorInfo instance and yield humar-readable
        lines one-by-one.

        Args:
            err: ErrorInfo instance

        Returns:
            Iterator yieldig formatted lines.
        """
        yield f"Error: {err.fingerprint}"
        yield self.get_exception_summary(err.exception)
        yield self.traceback_message()
        for fi in self.iter_stack(err):
            yield self.SEP
            if fi.source:
                yield (
                    f"File: {fi.source.file_name} "
                    f"(line {fi.source.current_line})"
                )
                for n, line in enumerate(
                    fi.source.lines, start=fi.source.first_line
                ):
                    sign = "==>" if n == fi.source.current_line else "   "
                    yield f"{n:5d} {sign} {line}"
                    if (
                        n == fi.source.current_line
                        and fi.source.pos
                        and fi.source.pos.start_line == fi.source.pos.end_line
                    ):
                        # Show caret
                        yield self.get_caret(line, fi.source.pos, 10)
            else:
                yield "File: <stdin> (line ???)"
            if fi.locals:
                yield "Locals:"
                for var_name, var_value in self.iter_vars(fi):
                    if len(var_value) > self.MAX_VAR_LEN:
                        yield f"{var_name:>20s} |\n{var_value}"
                    else:
                        yield f"{var_name:>20s} = {var_value}"
        yield self.SEP

    def iter_vars(
        self: "ExtendFormatter", fi: FrameInfo
    ) -> Iterable[Tuple[str, str]]:
        """Iterate frame variables and convert them to the readable form.

        Args:
            fi: FrameInfo instance

        Returns:
            Iterable of (`var name`, `var value`).
        """
        for k, v in fi.locals.items():
            try:
                rv = repr(v)
                if len(rv) > self.MAX_VAR_LEN:
                    rv = pformat(v)
            except Exception as e:  # noqa: BLE001
                rv = f"repr() failed: {e}"
            yield k, rv



# ---------------------------------------------------------------------
# Gufo Err: NeverFailFast
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""NeverFailFast."""

# Python modules
from types import TracebackType
from typing import Type

# Gufo Labs modules
from ..abc.failfast import BaseFailFast


class NeverFailFast(BaseFailFast):
    """Never fail-fast.

    Always returns False, so never inflicts fail-fast.

    Examples:
        ``` py
        err.setup(fail_fast=[NeverFailFast()])
        ```
    """

    def must_die(
        self: "NeverFailFast",
        t: Type[BaseException],
        v: BaseException,
        tb: TracebackType,
    ) -> bool:
        """Check if the process must die quickly.

        Always returns False.
        """
        return False



# ---------------------------------------------------------------------
# Gufo Err: ErrorInfoMiddleware
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""Compressor."""

# Python modules
import os
from typing import Callable, Dict, Optional, Tuple, Type


class Compressor(object):
    """Compressor/decompressor class.

    Use .encode() to compress data and .decode() to decompress.

    Args:
        format: Compression algorithm. One of:

            * `None` - do not compress
            * `gz` - GZip
            * `bz2` - BZip2
            * `xz` - LZMA/xz

    Raises:
        ValueError: If format is not supported.
    """

    FORMATS: Dict[
        Optional[str],
        Tuple[Callable[[bytes], bytes], Callable[[bytes], bytes]],
    ]

    def __init__(self: "Compressor", format: Optional[str] = None) -> None:
        try:
            self.encode, self.decode = self.FORMATS[format]
        except KeyError as e:
            msg = f"Unsupported format: {format}"
            raise ValueError(msg) from e
        if format is None:
            self.suffix = ""
        else:
            self.suffix = f".{format}"

    @classmethod
    def autodetect(cls: Type["Compressor"], path: str) -> "Compressor":
        """Returns Compressor instance for given format.

        Args:
            path: File path

        Returns:
            Compressor instance
        """
        return Compressor(format=cls.get_format(path))

    @classmethod
    def get_format(cls: Type["Compressor"], path: str) -> Optional[str]:
        """Auto-detect format from path.

        Args:
            path: File path.

        Returns:
            `format` parameter.
        """
        _, ext = os.path.splitext(path)
        if ext.startswith("."):
            fmt = ext[1:]
            if fmt in cls.FORMATS:
                return fmt
        return None

    @staticmethod
    def encode_none(data: bytes) -> bytes:
        """Encoder for `none` format.

        Args:
            data: Input bytes

        Returns:
            data as is.
        """
        return data

    @staticmethod
    def decode_none(data: bytes) -> bytes:
        """Decoder for `none` format.

        Args:
            data: Input bytes

        Returns:
            data as is.
        """
        return data

    @staticmethod
    def encode_gz(data: bytes) -> bytes:
        """Encoder for `gz` format.

        Args:
            data: Input bytes

        Returns:
            gzipped stream as bytes.
        """
        import gzip

        return gzip.compress(data)

    @staticmethod
    def decode_gz(data: bytes) -> bytes:
        """Decoder for `gz` format.

        Args:
            data: gzipped data as bytes.

        Returns:
            Ucompressed bytes.
        """
        import gzip

        return gzip.decompress(data)

    @staticmethod
    def encode_bz2(data: bytes) -> bytes:
        """Encoder for `bz2` format.

        Args:
            data: Input bytes

        Returns:
            bzipped stream as bytes.
        """
        import bz2

        return bz2.compress(data)

    @staticmethod
    def decode_bz2(data: bytes) -> bytes:
        """Encoder for `bz2` format.

        Args:
            data: bzipped data as bytes.

        Returns:
            Ucompressed bytes.
        """
        import bz2

        return bz2.decompress(data)

    @staticmethod
    def encode_xz(data: bytes) -> bytes:
        """Encoder for `xz` format.

        Args:
            data: Input bytes

        Returns:
            xzipped stream as bytes.
        """
        import lzma

        return lzma.compress(data)

    @staticmethod
    def decode_xz(data: bytes) -> bytes:
        """Decoder for `xz` format.

        Args:
            data: xzipped data as bytes.

        Returns:
            Ucompressed bytes.
        """
        import lzma

        return lzma.decompress(data)


Compressor.FORMATS = {
    None: (Compressor.encode_none, Compressor.decode_none),
    "gz": (Compressor.encode_gz, Compressor.decode_gz),
    "bz2": (Compressor.encode_bz2, Compressor.decode_bz2),
    "xz": (Compressor.encode_xz, Compressor.decode_xz),
}



# ---------------------------------------------------------------------
# Gufo Err: abc module
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""Abstract base classes to be used to extend the functionality.

* [FailFast][gufo.err.abc.failfast.BaseFailFast]:
  Fail-fast processing.
* [Formatter][gufo.err.abc.formatter.BaseFormatter]:
  Traceback formatters.
* [Middleware][gufo.err.abc.middleware.BaseMiddleware]:
  Error-processing middleware.
"""



# ---------------------------------------------------------------------
# Gufo Err: BaseFailFast method
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------
"""BaseFailFast."""

# Python modules
from abc import ABC, abstractmethod
from types import TracebackType
from typing import Type


class BaseFailFast(ABC):
    """Abstract base type for fail-fast behavior.

    Fail-fast classes must implement `must_die` method.
    When fail-fast check decides the error is unrecoverable,
    it must return `True` value.
    """

    @abstractmethod
    def must_die(
        self: "BaseFailFast",
        t: Type[BaseException],
        v: BaseException,
        tb: TracebackType,
    ) -> bool:
        """Fail-fast check. Must be overriden in subclasses.

        Args:
            t: Exception type. Same as `sys.exc_info()[0]`.
            v: Exception value. Same as `sys.exc_info()[1]`.
            tb: Traceback. Same as `sys.exc_info()[2]`.

        Returns:
            * `True`, if the error is not recoverable and the process
            must be terminated ASAP.
            * `False` to pass to the next check.
        """
        ...
